from __future__ import print_function

import sys

import numpy as np
from pyteomics.xml import cvstr


def _log(message):
    print(message, file=sys.stderr)


class LoggingProxy(object):
    def __init__(self, logger=None):
        self.logger = logger

    def enable(self, logger=None):
        if logger is None:
            logger = _log
        self.logger = logger

    def log(self, message):
        if self.logger is not None:
            self.logger(message)

    def __call__(self, message):
        self.log(message)

    def disable(self):
        self.logger = None


log = LoggingProxy()
log.enable()


class TransformerBase(object):
    def log(self, *message):
        log(', '.join(map(str, message)))

    def __enter__(self):
        return self.writer.__enter__()

    def __exit__(self, *args, **kwargs):
        self.writer.__exit__(*args, **kwargs)


def key_fn(x):
    return (getattr(x, 'accession', None), str(x))


def differ(a, b):
    if not issubclass(type(a), type(b)):
        return False
    if isinstance(a, dict):
        return dict_diff(a, b)
    elif isinstance(a, (list, tuple)):
        return seq_diff(a, b)
    elif isinstance(a, float):
        return abs(a - b) < 1e-2
    elif isinstance(a, cvstr):
        if a.accession is not None:
            return a.accession.lower() == b.accession.lower()
        else:
            return a == b
    elif isinstance(a, np.ndarray):
        return np.allclose(a, b)
    else:
        return a == b


def seq_diff(source, test):
    source = list(source)
    test = list(test)
    if len(source) != len(test):
        return False
    for s in source:
        for i, t in enumerate(list(test)):
            if differ(s, t):
                test.pop(i)
                break
    return len(test) == 0


def dict_diff(a, b):
    a = dict(a)
    b = dict(b)
    aparams = []
    for key, value in list(a.items()):
        if hasattr(key, 'accession'):
            a.pop(key)
            aparams.append((key, value))
    bparams = []
    for key, value in list(b.items()):
        if hasattr(key, 'accession'):
            b.pop(key)
            bparams.append((key, value))
    aparams.sort()
    bparams.sort()
    if sorted(a.keys(), key=str) != sorted(b.keys(), key=str):
        return False
    for akey, avalue in a.items():
        for bkey, bvalue in b.items():
            if akey == bkey:
                if differ(avalue, bvalue):
                    pass
                else:
                    print(type(avalue))
                    print(akey, avalue, '!=', bvalue)
                    return False
                break

    param_diff = seq_diff(aparams, bparams)
    if not param_diff:
        print("Parameters Differ")
    return param_diff



import pkg_resources

from lxml import etree

from six import raise_from


def get_xsd(name):
    return pkg_resources.resource_stream(__name__, "xsd/%s" % name)


schemas = {
    'mzML': ('mzML1.1.0.xsd'),
    'indexedmzML': ('mzML1.1.2_idx.xsd'),
    'MzIdentML': ('mzIdentML1.2.0.xsd'),
    "http://psidev.info/psi/pi/mzIdentML/1.1 ../schema/mzIdentML1.1.0.xsd": ("mzIdentML1.1.0.xsd"),
    "http://psidev.info/psi/pi/mzIdentML/1.1.1 ../schema/mzIdentML1.1.1.xsd": ("mzIdentML1.1.1.xsd"),
    "http://psidev.info/psi/pi/mzIdentML/1.2 ../schema/mzIdentML1.2.0.xsd": ("mzIdentML1.2.0.xsd"),
}


def get_schema(name):
    schema_name = schemas[name]
    tree = etree.parse(get_xsd(schema_name))
    return etree.XMLSchema(tree)


def validate(path):
    tree = etree.parse(path)
    root = tree.getroot()

    location = root.attrib.get('{http://www.w3.org/2001/XMLSchema-instance}schemaLocation')

    try:
        schema = get_schema(location)
    except KeyError:
        try:
            parts = root.tag.split("}", 1)
            if len(parts) == 1:
                name = parts[0]
            else:
                name = parts[1]
            schema = get_schema(name)
        except KeyError:
            raise_from(KeyError("Could not locate a schema for %r or %r" % (name, location)), None)
    result = schema.validate(tree)
    return result, schema



# pylint: disable=super-init-not-called
import os
import warnings
try:
    from collections.abc import Mapping, Iterable
except ImportError:
    from collections import Mapping, Iterable
from numbers import Number
from ..utils import checksum_file
from ..xml import _element, element, TagBase, CV
from ..document import (
    ComponentBase as _ComponentBase,
    DocumentContext,
    XMLBindingDispatcherBase,
    ParameterContainer,
    IDParameterContainer)
from .binary_encoding import dtype_to_encoding, compression_map, encode_array
from .utils import ensure_iterable, basestring


try:
    FileNotFoundError  # pylint: disable=used-before-assignment
except Exception as e:
    FileNotFoundError = OSError


class MzML(TagBase):
    type_attrs = {
        'xmlns': 'http://psi.hupo.org/ms/mzml',
        'version': '1.1.0',
        'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xsi:schemaLocation': 'http://psi.hupo.org/ms/mzml http://psidev.info/files/ms/mzML/xsd/mzML1.1.0.xsd'}

    def __init__(self, **attrs):
        super(MzML, self).__init__('mzML', **attrs)


_COMPONENT_NAMESPACE = 'mzml'
_xmlns = 'http://psidev.info/psi/pi/mzML/1.1'


class IndexedMzML(TagBase):
    type_attrs = {
        "xmlns": "http://psi.hupo.org/ms/mzml",
        'xmlns:xsi': "http://www.w3.org/2001/XMLSchema-instance",
        'xsi:schemaLocation': "http://psi.hupo.org/ms/mzml http://psidev.info/files/ms/mzML/xsd/mzML1.1.2_idx.xsd"
    }

    def __init__(self, **attrs):
        super(IndexedMzML, self).__init__('indexedmzML', **attrs)


class ComponentDispatcher(XMLBindingDispatcherBase):

    def __init__(self, *args, **kwargs):
        super(
            ComponentDispatcher,
            self).__init__(
            component_namespace=_COMPONENT_NAMESPACE,
            *args,
            **kwargs)


class ComponentBase(_ComponentBase):
    component_namespace = _COMPONENT_NAMESPACE


default_cv_list = [
    CV(id='PSI-MS',
       uri='https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo',
       full_name='PSI-MS'),
    CV(id='UO',
       uri='http://ontologies.berkeleybop.org/uo.obo',
       full_name='UNIT-ONTOLOGY')
]


NullMap = DocumentContext(default_cv_list)


class GenericCollection(ComponentBase):
    requires_id = False

    def __init__(self, tag_name, members, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, count=len(self.members))

    def write_content(self, xml_file):
        for member in self.members:
            member.write(xml_file)

    def __len__(self):
        return len(self.members)

    def __iter__(self):
        return iter(self.members)

    def __getitem__(self, i):
        return self.members[i]


class IDGenericCollection(GenericCollection):

    def __init__(self, tag_name, members, id, context=NullMap):
        self.context = context
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, id=id, count=len(self.members))
        self.context[tag_name][id] = self.element.id

    def write_content(self, xml_file):
        for member in self.members:
            member.write(xml_file)


class FileContent(ComponentBase):
    requires_id = False

    def __init__(self, params=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.element = _element('fileContent')
        self.context = context

    def write_content(self, xml_file):
        self.write_params(xml_file)


class SourceFileList(GenericCollection):

    def __init__(self, members, context):
        super(SourceFileList, self).__init__(
            'sourceFileList', members, context)


class SourceFile(ComponentBase):
    requires_id = True

    def __init__(self, location=None, name=None, id=None,
                 params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        self.context = context
        self.params = self.prepare_params(params, **kwargs)
        self.location = location
        self.name = name
        self.element = _element(
            'sourceFile',
            location=location,
            id=id,
            name=name)
        context['SourceFile'][id] = self.element.id

    @property
    def path(self):
        return os.path.join([self.location, self.name])

    def is_local(self):
        path = self.path
        if path.startswith("file:///"):
            path = path.replace("file:///", '', 1)
        return os.path.exists(path)

    def checksum(self, digest='sha-1'):
        if not self.is_local():
            raise FileNotFoundError("Can't checksum %r, file not found")
        path = self.path
        if path.startswith("file:///"):
            path = path.replace("file:///", '', 1)
        checksum = checksum_file(path, digest)
        return self.context.param(name=digest, value=checksum)

    def write_content(self, xml_file):
        self.write_params(xml_file)


class FileDescription(ComponentBase):
    requires_id = False

    def __init__(self, content, source_files, contacts=None, context=NullMap):
        if (not isinstance(content, FileContent)):
            content = FileContent(content, context=context)
        if (not isinstance(source_files, SourceFileList)):
            if ((len(source_files) > 0) and (
                    not isinstance(source_files[0], SourceFile))):
                source_files = [SourceFile(context=context, **f)
                                for f in source_files]
            source_files = SourceFileList(source_files, context=context)
        contacts = ensure_iterable(contacts)
        if (contacts and (not isinstance(contacts[0], Contact))):
            contacts = [Contact.ensure(contact, context=context)
                        for contact in contacts]
        self.content = content
        self.source_files = source_files
        self.contacts = contacts
        self.context = context
        self.element = _element('fileDescription')

    def write_content(self, xml_file):
        self.content.write(xml_file)
        self.source_files.write(xml_file)
        for contact in self.contacts:
            contact.write(xml_file)


class ReferenceableParamGroupList(GenericCollection):

    def __init__(self, members, context):
        super(ReferenceableParamGroupList, self).__init__(
            'referenceableParamGroupList', members, context)


class ReferenceableParamGroup(IDParameterContainer):

    def __init__(self, params=None, id=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        super(
            ReferenceableParamGroup,
            self).__init__(
            'referenceableParamGroup',
            params,
            dict(
                id=id),
            context=context,
            **kwargs)
        self.id = self.element.id
        self.context = context
        self.context['ReferenceableParamGroup'][id] = self.element.id


class SampleList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(SampleList, self).__init__('sampleList', members, context)


class Sample(IDParameterContainer):

    def __init__(self, name, params=None, id=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        params.extend(kwargs.items())
        super(
            Sample,
            self).__init__(
            'sample',
            params,
            dict(
                name=name,
                id=id),
            context)
        self.name = name
        self.id = self.element.id
        context['Sample'][id] = self.element.id


class SoftwareList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(SoftwareList, self).__init__('softwareList', members, context)


class Software(IDParameterContainer):

    def __init__(self, id=None, version='0.0',
                 params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        params.extend(kwargs.items())
        super(
            Software,
            self).__init__(
            'software',
            params,
            dict(
                id=id,
                version=version),
            context=context)
        self.version = version
        self.id = self.element.id
        context['Software'][id] = self.element.id

    @property
    def with_id(self):
        return True


class ScanSettingsList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(
            ScanSettingsList,
            self).__init__(
            'scanSettingsList',
            members,
            context)


class ScanSettings(ComponentBase):
    requires_id = True

    def __init__(self, id, source_file_references=None, target_list=None, params=None, context=NullMap, **kwargs):
        if target_list is None:
            target_list = []
        if not isinstance(target_list, TargetList):
            target_list = TargetList(ensure_iterable(target_list), context=context)
        self.context = context
        self.params = self.prepare_params(params, **kwargs)
        self.source_file_references = source_file_references
        self.target_list = target_list
        self.element = _element('scanSettings', id=id)
        self.context['ScanSettings'][id] = self.element.id

    def _prepare_source_file_reference(self):
        temp = []
        for ref in ensure_iterable(self.source_file_references):
            if isinstance(ref, Mapping):
                temp.append(SourceFileRef.ensure(ref, context=self.context))
            elif isinstance(ref, SourceFileRef):
                temp.append(ref)
            else:
                temp.append(SourceFileRef(ref, context=self.context))
        self.source_file_references = temp

    def write_content(self, xml_file):
        source_refs = self.source_file_references
        if len(source_refs) > 0:
            with xml_file.element("sourceFileRefList", count=len(source_refs)):
                for ref in source_refs:
                    xml_file.write(ref.element())
        if self.target_list:
            self.target_list.write(xml_file)
        self.write_params(xml_file)


class SourceFileRef(ComponentBase):
    requires_id = False

    def __init__(self, source_file_id, context=NullMap):
        self.source_file_id = source_file_id
        self.context = context
        self.ref = self.context['SourceFile'][self.source_file_id]
        self.element = _element("sourceFileRef", ref=self.ref)

    def write_content(self, xml_file):
        pass


class Target(ParameterContainer):
    requires_id = False

    def __init__(self, params, context=NullMap, **kwargs):
        if (params is None):
            params = []
        params = self.prepare_params(params, **kwargs)
        super(Target, self).__init__('target', params, context=context)


class TargetList(GenericCollection):
    def __init__(self, members, context=NullMap):
        members = [Target.ensure(t, context=context) for t in members]
        super(TargetList, self).__init__(
            'targetList', members, context)


class InstrumentConfigurationList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(InstrumentConfigurationList, self).__init__(
            'instrumentConfigurationList', members, context)


class InstrumentConfiguration(ComponentBase):
    requires_id = True

    def __init__(self, id, component_list=None, params=None,
                 software_reference=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        if (not isinstance(component_list, ComponentList)):
            component_list = ComponentList(component_list, context=context)
        self.params = self.prepare_params(params, **kwargs)
        self.software_reference = software_reference
        if software_reference is not None:
            self._software_reference = context['Software'][software_reference]
        self.component_list = component_list
        self.element = _element('instrumentConfiguration', id=id)
        self.context = context
        self.context['InstrumentConfiguration'][id] = self.element.id

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if (self.component_list is not None):
            self.component_list.write(xml_file)
        if (self.software_reference is not None):
            _element(
                'softwareRef',
                ref=self._software_reference).write(xml_file)


class ComponentList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(ComponentList, self).__init__('componentList', members, context)

    @classmethod
    def build(cls, members, context=NullMap, type_key='type'):
        components = []
        for component in sorted(members, key=(lambda x: int(x['order']))):
            if (component[type_key] == 'source'):
                components.append(
                    Source(int(component['order']), component['params'], context))
            elif (component[type_key] == 'analyzer'):
                components.append(
                    Analyzer(int(component['order']), component['params'], context))
            elif (component[type_key] == 'detector'):
                components.append(
                    Detector(int(component['order']), component['params'], context))
            else:
                raise KeyError(('Unknown component %s' % component[type_key]))
        return cls(components, context=context)


class Source(ParameterContainer):

    def __init__(self, order, params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        params.extend(kwargs.items())
        super(
            Source,
            self).__init__(
            'source',
            params,
            dict(
                order=order),
            context=context)
        try:
            self.order = int(order)
        except (ValueError, TypeError):
            self.order = order


class Analyzer(ParameterContainer):

    def __init__(self, order, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        super(
            Analyzer,
            self).__init__(
            'analyzer',
            params,
            dict(
                order=order),
            context=context)
        try:
            self.order = int(order)
        except (ValueError, TypeError):
            self.order = order


class Detector(ParameterContainer):

    def __init__(self, order, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        super(
            Detector,
            self).__init__(
            'detector',
            params,
            dict(
                order=order),
            context=context)
        try:
            self.order = int(order)
        except (ValueError, TypeError):
            self.order = order


class DataProcessingList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(
            DataProcessingList,
            self).__init__(
            'dataProcessingList',
            members,
            context)


class DataProcessing(ComponentBase):
    requires_id = True

    def __init__(self, processing_methods=None, id=None, context=NullMap):
        if (processing_methods is None):
            processing_methods = []
        if (processing_methods and (not isinstance(
                processing_methods[0], ProcessingMethod))):
            processing_methods = [
                ProcessingMethod(
                    context=context,
                    **m) for m in processing_methods]
        self.processing_methods = processing_methods
        self.element = _element('dataProcessing', id=id)
        self.context = context
        context['DataProcessing'][id] = self.element.id

    def write_content(self, xml_file):
        for method in self.processing_methods:
            method.write(xml_file)


class ProcessingMethod(ParameterContainer):

    def __init__(self, order, software_reference,
                 params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.order = order
        self.software_reference = software_reference
        self._software_reference = context['Software'][software_reference]
        self.element = _element(
            'processingMethod',
            order=order,
            softwareRef=self._software_reference)
        self.params = params
        self.context = context


class SpectrumList(ComponentBase):
    requires_id = False

    def __init__(self, members, default_data_processing_reference,
                 context=NullMap):
        self.members = members
        self.default_data_processing_reference = default_data_processing_reference
        self._default_data_processing_reference = context[
            'DataProcessing'][default_data_processing_reference]
        self.element = _element(
            'spectrumList',
            count=len(
                self.members),
            defaultDataProcessingRef=self._default_data_processing_reference)
        self.context = context

    def write_content(self, xml_file):
        for member in self.members:
            member.write(xml_file)


class Spectrum(ComponentBase):
    requires_id = True

    def __init__(self, index, binary_data_list=None, scan_list=None, precursor_list=None, product_list=None,
                 default_array_length=None, source_file_reference=None, data_processing_reference=None, id=None,
                 params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        self.index = index
        self.binary_data_list = binary_data_list
        self.scan_list = scan_list
        self.data_processing_reference = data_processing_reference
        if data_processing_reference is not None:
            self._data_processing_reference = context['DataProcessing'][data_processing_reference]
        else:
            self._data_processing_reference = None
        self.precursor_list = precursor_list
        self.product_list = product_list
        self.default_array_length = default_array_length
        self.source_file_reference = source_file_reference
        if source_file_reference is not None:
            self._source_file_reference = context['SourceFile'][source_file_reference]
        else:
            self._source_file_reference = None
        self.element = _element(
            'spectrum',
            id=id,
            index=index,
            sourceFileRef=self._source_file_reference,
            defaultArrayLength=self.default_array_length,
            dataProcessingRef=self._data_processing_reference)
        self.context = context
        self.context['Spectrum'][id] = self.element.id
        self.params = self.prepare_params(params, **kwargs)
        self._check_params()

    def _check_params(self):
        ms_level = None
        spectrum_type = None
        for param in self.params:
            param = self.context.param(param)
            try:
                term = self.context.term(param.accession)
            except (AttributeError, KeyError):
                continue
            if (term.id == 'MS:1000511'):
                ms_level = int(param.value)
            elif ((term.id == 'MS:1000579') or (term.id == 'MS:1000580')):
                spectrum_type = term.name
        if ((ms_level is not None) and (spectrum_type is None)):
            spectrum_type = ('MS:1000579' if (ms_level == 1) else 'MS:1000580')
            self.params.append(self.context.param(spectrum_type))
        elif ((spectrum_type is not None) and (ms_level is None)):
            if (spectrum_type == 'MS:1000579'):
                self.params.append(
                    self.context.param(
                        accession='MS:1000511',
                        value=1))
            else:
                raise ValueError(
                    "A spectrum without MS:100511 'ms level' and cannot be determined from other parameters")

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if (self.scan_list is not None):
            self.scan_list.write(xml_file)
        if (self.precursor_list is not None):
            self.precursor_list.write(xml_file)
        if (self.product_list is not None):
            self.product_list.write(xml_file)
        self.binary_data_list.write(xml_file)


class Run(ComponentBase):
    requires_id = True

    def __init__(self, default_instrument_configuration_reference, spectrum_list=None, chromatogram_list=None, id=None,
                 default_source_file_reference=None, sample_reference=None, start_time_stamp=None, params=None,
                 context=NullMap):
        if (params is None):
            params = []
        if (spectrum_list is None):
            spectrum_list = []
        if (chromatogram_list is None):
            chromatogram_list = []
        self.params = params
        self.spectrum_list = spectrum_list
        self.chromatogram_list = chromatogram_list
        self.sample_reference = sample_reference
        self._sample_reference = context['Sample'][sample_reference]
        self.default_instrument_configuration_reference = default_instrument_configuration_reference
        self._default_instrument_configuration_reference = context[
            'InstrumentConfiguration'][default_instrument_configuration_reference]
        self.default_source_file_reference = default_source_file_reference
        self._default_source_file_reference = context[
            'SourceFile'][default_source_file_reference]
        self.start_time_stamp = start_time_stamp
        self.element = _element(
            'run',
            id=id,
            defaultInstrumentConfigurationRef=self._default_instrument_configuration_reference,
            defaultSourceFileRef=self._default_source_file_reference,
            sampleRef=self._sample_reference,
            startTimeStamp=start_time_stamp)
        self.context = context

    def write_content(self, xml_file):
        self.write_params(xml_file)


class BinaryDataArrayList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(BinaryDataArrayList, self).__init__(
            'binaryDataArrayList', members, context)


class BinaryDataArray(ComponentBase):
    requires_id = False

    def __init__(self, binary, encoded_length, data_processing_reference=None,
                 array_length=None, params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        self.encoded_length = encoded_length
        self.data_processing_reference = data_processing_reference
        if data_processing_reference:
            self._data_processing_reference = context[
                'DataProcessing'][data_processing_reference]
        else:
            self._data_processing_reference = None
        self.array_length = array_length
        self.params = self.prepare_params(params, **kwargs)
        self.binary = binary
        self.element = _element(
            'binaryDataArray',
            arrayLength=array_length,
            encodedLength=encoded_length,
            dataProcessingRef=self._data_processing_reference)
        self.context = context
        self._array_type = None

    def _find_array_type(self):
        for param in self.params:
            param = self.context.param(param)
            if param.accession is None:
                continue
            try:
                term = self.context.term(param.accession)
            except KeyError:
                continue
            if term.is_of_type('MS:1000513'):
                return term

    @property
    def array_type(self):
        if self._array_type is None:
            self._array_type = self._find_array_type()
        return self._array_type

    @array_type.setter
    def array_type(self, value):
        self._array_type = value

    def write_content(self, xml_file):
        self.write_params(xml_file)
        self.binary.write(xml_file)

    @classmethod
    def from_array(cls, data_array, compression=None, data_processing_reference=None, params=None, context=None):
        if (params is None):
            params = []
        if (context is None):
            context = NullMap
        compression = compression_map[compression]
        encoded_binary = encode_array(
            data_array,
            compression=compression,
            dtype=data_array.dtype.type)
        binary = Binary(encoded_binary)
        array_length = len(data_array)
        params.append(compression_map[compression])
        params.append(dtype_to_encoding[data_array.dtype.type])
        encoded_length = len(encoded_binary)
        inst = cls(
            binary,
            encoded_length,
            array_length=array_length,
            data_processing_reference=data_processing_reference,
            params=params,
            context=context)
        return inst


class Binary(ComponentBase):
    requires_id = False

    def __init__(self, encoded_array, context=NullMap):
        self.encoded_array = encoded_array
        self.context = context
        self.element = _element('binary', text=encoded_array)

    def write_content(self, xml_file):
        xml_file.write(self.encoded_array)


class ScanList(ComponentBase):
    requires_id = False

    def __init__(self, members, params=None, context=NullMap):
        if (params is None):
            params = []
        self.members = members
        self.params = params
        self.element = _element('scanList', count=len(self.members))
        self.context = context

    def write_content(self, xml_file):
        self.write_params(xml_file)
        for member in self.members:
            member.write(xml_file)


class Scan(ComponentBase):
    requires_id = False

    def __init__(self, scan_window_list=None, instrument_configuration_ref=None,
                 source_file_reference=None, spectrum_reference=None, external_spectrum_id=None,
                 params=None, context=NullMap, **kwargs):
        if (scan_window_list is None):
            scan_window_list = ScanWindowList([], context)
        elif (not isinstance(scan_window_list, ScanWindowList)):
            scan_window_list = ScanWindowList(scan_window_list, context)
        self.instrument_configuration_ref = instrument_configuration_ref
        self._instrument_configuration_ref = context['InstrumentConfiguration'].get(
            instrument_configuration_ref)
        self.source_file_reference = source_file_reference
        self._source_file_reference = None
        if self.source_file_reference:
            self._source_file_reference = context['SourceFile'][self.source_file_reference]

        # these references make no sense given the document structure but they're in the spec?
        self.spectrum_reference = spectrum_reference
        self.external_spectrum_id = external_spectrum_id

        self.params = self.prepare_params(params, **kwargs)
        self.scan_window_list = scan_window_list
        self.element = _element(
            'scan', instrumentConfigurationRef=self._instrument_configuration_ref,
            externalSpectrumID=self.external_spectrum_id, sourceFileRef=self._source_file_reference,
            spectrumRef=self.spectrum_reference)
        self.context = context

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if (len(self.scan_window_list) > 0):
            self.scan_window_list.write(xml_file)


class ScanWindowList(GenericCollection):

    def __init__(self, members, context=NullMap):
        components = []
        for member in members:
            if isinstance(member, (list, tuple)):
                components.append(ScanWindow(context=context, *member))
            elif isinstance(member, Mapping):
                components.append(ScanWindow(context=context, **member))
            else:
                components.append(member)
        super(
            ScanWindowList,
            self).__init__(
            'scanWindowList',
            components,
            context=context)


class ScanWindow(ComponentBase):
    requires_id = False

    def __init__(self, lower, upper, params=None, context=NullMap):
        if (params is None):
            params = []
        self.lower = lower
        self.upper = upper
        self.params = params
        self.element = _element('scanWindow')
        self.context = context

    def write_content(self, xml_file):
        self.context.param(
            name='scan window lower limit',
            value=self.lower,
            unit_name='m/z',
            unit_accession='MS:1000040',
            unit_cv_ref='MS')(xml_file)
        self.context.param(
            name='scan window upper limit',
            value=self.upper,
            unit_name='m/z',
            unit_accession='MS:1000040',
            unit_cv_ref='MS')(xml_file)
        self.write_params(xml_file)


class IsolationWindow(ComponentBase):
    requires_id = False

    def __init__(self, lower=None, target=None, upper=None, params=None, context=NullMap, **kwargs):
        if (params is None):
            params = []
        self.target = target
        self.lower = lower
        self.upper = upper
        self.element = _element('isolationWindow')
        self.context = context
        self.params = self.prepare_params(params, **kwargs)

    def write_content(self, xml_file):
        if self.target is not None:
            self.context.param(
                name='isolation window target m/z',
                value=self.target,
                unit_name='m/z',
                unit_accession='MS:1000040',
                unit_cv_ref='MS')(xml_file)
        if self.lower is not None:
            self.context.param(
                name='isolation window lower offset',
                value=self.lower,
                unit_name='m/z',
                unit_accession='MS:1000040',
                unit_cv_ref='MS')(xml_file)
        if self.upper is not None:
            self.context.param(
                name='isolation window upper offset',
                value=self.upper,
                unit_name='m/z',
                unit_accession='MS:1000040',
                unit_cv_ref='MS')(xml_file)
        self.write_params(xml_file)


class PrecursorList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(
            PrecursorList,
            self).__init__(
            'precursorList',
            members,
            context=context)


class Precursor(ComponentBase):
    requires_id = False

    def __init__(self, selected_ion_list, activation, isolation_window=None, spectrum_reference=None,
                 source_file_reference=None, external_spectrum_id=None, context=NullMap):
        if (isolation_window is not None):
            if isinstance(isolation_window, (tuple, list)):
                isolation_window = IsolationWindow(
                    context=context, *isolation_window)
            elif isinstance(isolation_window, Mapping):
                isolation_window = IsolationWindow(
                    context=context, **isolation_window)
        if selected_ion_list is not None:
            selected_ion_list = SelectedIonList(
                [SelectedIon.ensure(si, context=context) for si in list(selected_ion_list)],
                context=context)
        self.selected_ion_list = selected_ion_list
        if activation is None:
            warnings.warn("Precursor element with missing activation information")
            activation = {}
        self.activation = Activation.ensure(activation, context=context)
        self.isolation_window = isolation_window

        self.spectrum_reference = spectrum_reference
        self._spectrum_reference = None
        if self.spectrum_reference:
            self._spectrum_reference = context['Spectrum'][spectrum_reference]

        self.source_file_reference = source_file_reference
        self._source_file_reference = None
        if self.source_file_reference:
            self._source_file_reference = context['SourceFile'][self.source_file_reference]

        self.external_spectrum_id = external_spectrum_id

        self.element = _element(
            'precursor', spectrumRef=self._spectrum_reference,
            sourceFileRef=self._source_file_reference,
            externalSpectrumID=self.external_spectrum_id)
        self.context = context

    def write_content(self, xml_file):
        if self.isolation_window is not None:
            self.isolation_window.write(xml_file)
        if self.selected_ion_list is not None and self.selected_ion_list:
            self.selected_ion_list.write(xml_file)
        if self.activation is not None:
            self.activation.write(xml_file)


class ProductList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(
            ProductList,
            self).__init__(
            'productList',
            members,
            context=context)


class Product(ComponentBase):
    requires_id = False

    def __init__(self, isolation_window=None,
                 spectrum_reference=None, context=NullMap):
        if (isolation_window is not None):
            if isinstance(isolation_window, (tuple, list)):
                isolation_window = IsolationWindow(
                    context=context, *isolation_window)
            elif isinstance(isolation_window, Mapping):
                isolation_window = IsolationWindow(
                    context=context, **isolation_window)
        self.isolation_window = isolation_window
        self.element = _element('product')
        self.context = context

    def write_content(self, xml_file):
        if (self.isolation_window is not None):
            self.isolation_window.write(xml_file)


class Activation(ParameterContainer):

    def __init__(self, params=None, context=NullMap, **kwargs):
        super(Activation, self).__init__('activation', params, context=context, **kwargs)


class SelectedIonList(GenericCollection):

    def __init__(self, members, context=NullMap):
        super(
            SelectedIonList,
            self).__init__(
            'selectedIonList',
            members,
            context=context)


class SelectedIon(ComponentBase):
    requires_id = False

    def __init__(self, selected_ion_mz, intensity=None,
                 charge=None, params=None,
                 intensity_unit='number of detector counts', context=NullMap):
        if (params is None):
            params = []
        self.selected_ion_mz = selected_ion_mz
        self.intensity = intensity
        self.charge = charge
        self.params = params
        self.element = _element('selectedIon')
        self.context = context
        self.intensity_unit = intensity_unit

    def write_content(self, xml_file):
        if (self.selected_ion_mz is not None):
            self.context.param(
                name='selected ion m/z',
                value=self.selected_ion_mz,
                unit_name='m/z')(xml_file)
        if (self.intensity is not None):
            self.context.param(
                name='peak intensity',
                value=self.intensity,
                unit_name=self.intensity_unit)(xml_file)
        if (self.charge is not None):
            if isinstance(self.charge, Number):
                self.context.param(
                    name='charge state', value=int(
                        self.charge))(xml_file)
            elif isinstance(self.charge, Iterable) and not isinstance(self.charge, basestring):
                for value in self.charge:
                    self.context.param(name='possible charge state', value=value)(xml_file)
            else:
                warnings.warn(
                    ('Invalid charge state provided (%r)' %
                     (self.charge,)))
        self.write_params(xml_file)


class Chromatogram(ComponentBase):
    requires_id = True

    def __init__(self, index, binary_data_list=None, precursor=None, product=None,
                 default_array_length=None, data_processing_reference=None, id=None, params=None, context=NullMap):
        if (params is None):
            params = []
        self.index = index
        self.default_array_length = default_array_length
        self.binary_data_list = binary_data_list
        self.precursor = precursor
        self.product = product
        self.data_processing_reference = data_processing_reference
        if data_processing_reference:
            self._data_processing_reference = context[
                'DataProcessing'][data_processing_reference]
        else:
            self._data_processing_reference = None
        self.element = _element(
            'chromatogram',
            id=id,
            index=index,
            defaultArrayLength=self.default_array_length,
            dataProcessingRef=self._data_processing_reference)
        self.context = context
        self.context['Chromatogram'][id] = self.element.id
        self.params = params

    def write_content(self, xml_file):
        self.write_params(xml_file)
        if (self.precursor is not None):
            self.precursor.write(xml_file)
        if (self.product is not None):
            self.product.write(xml_file)
        self.binary_data_list.write(xml_file)


DEFAULT_CONTACT_ID = 'PERSON_DOC_OWNER'
DEFAULT_ORGANIZATION_ID = 'ORG_DOC_OWNER'


class CVList(ComponentBase):
    requires_id = False

    def __init__(self, cv_list=None, context=NullMap):
        if (cv_list is None):
            cv_list = default_cv_list
        self.cv_list = cv_list
        self.element = _element("cvList", count=len(self.cv_list))

    def write_content(self, xml_file):
        for member in self.cv_list:
            tag = _element(
                'cv',
                id=member.id,
                fullName=member.full_name,
                URI=member.uri)
            if (member.version is not None):
                tag.attrs['version'] = member.version
            if member.options:
                tag.attrs.update(member.options)
            xml_file.write(tag.element(with_id=True))

    def __iter__(self):
        return iter(self.cv_list)


class Contact(ParameterContainer):

    def __init__(self, params, context=NullMap, **kwargs):
        super(
            Contact,
            self).__init__(
            'contact',
            params,
            context=context,
            **kwargs)


class Person(ComponentBase):
    requires_id = True

    def __init__(self, first_name='first_name', last_name='last_name',
                 id=DEFAULT_CONTACT_ID, affiliation=DEFAULT_ORGANIZATION_ID, context=NullMap):
        self.first_name = first_name
        self.last_name = last_name
        self.id = id
        self.affiliation = affiliation
        self.element = _element(
            'Person',
            firstName=first_name,
            last_name=last_name,
            id=id)
        context['Person'][id] = self.element.id
        self.context = context

    def write_content(self, xml_file):
        element(xml_file, 'Affiliation', organization_ref=self.affiliation)


class Organization(ComponentBase):
    requires_id = True

    def __init__(self, name='name', id=DEFAULT_ORGANIZATION_ID,
                 context=NullMap):
        self.name = name
        self.id = id
        self.element = _element('Organization', name=name, id=id)
        context['Organization'][id] = self.id
        self.context = context

    def write_content(self, xml_file):
        xml_file.write(self.element.element())



import re

import io
from io import BytesIO
from collections import defaultdict, OrderedDict

from six import string_types as basestring, PY2

try:
    from collections.abc import Sequence, Mapping
except ImportError:
    from collections import Sequence, Mapping

from hashlib import sha1

from lxml import etree

from psims import compression


class Offset(object):
    __slots__ = ['offset', 'attrs']

    def __init__(self, offset, attrs):
        self.offset = offset
        self.attrs = attrs

    def __eq__(self, other):
        return int(self) == int(other)

    def __ne__(self, other):
        return int(self) != int(other)

    def __hash__(self):
        return hash(self.offset)

    def __int__(self):
        return self.offset

    def __index__(self):
        return self.offset

    def __repr__(self):
        template = "{self.__class__.__name__}({self.offset}, {self.attrs})"
        return template.format(self=self)


class TagIndexerBase(object):
    attr_pattern = re.compile(br"(\S+)=[\"']([^\"']+)[\"']")

    def __init__(self, name, pattern):
        if isinstance(pattern, str):
            pattern = pattern.encode("utf8")
        if isinstance(pattern, bytes):
            pattern = re.compile(pattern)
        self.name = name
        self.pattern = pattern
        self.index = OrderedDict()

    def __len__(self):
        return len(self.index)

    def __iter__(self):
        return iter(self.index.items())

    def scan(self, data, distance):
        is_match = self.pattern.search(data)
        if is_match:
            attrs = dict(self.attr_pattern.findall(data))
            xid = attrs[b'id']
            offset = Offset(distance + is_match.start(), attrs)
            self.index[xid] = offset
        return bool(is_match)

    def __call__(self, data, distance):
        return self.scan(data, distance)

    def write(self, writer):
        writer.write("    <index name=\"{}\">\n".format(self.name).encode('utf-8'))
        for ref_id, index_data in self.index.items():
            writer.write('      <offset idRef="{}">{:d}</offset>\n'.format(
                ref_id, int(index_data)).encode('utf-8'))
        writer.write(b"    </index>\n")

    def write_xml(self, writer):
        with writer.element("index", name=self.name):
            for ref_id, index_data in self.index.items():
                try:
                    ref_id = ref_id.decode("utf8")
                except AttributeError:
                    pass
                with writer.element("offset", idRef=ref_id):
                    writer.write(str(int(index_data)))


class SpectrumIndexer(TagIndexerBase):
    def __init__(self):
        super(SpectrumIndexer, self).__init__(
            'spectrum', re.compile(b"<spectrum "))


class ChromatogramIndexer(TagIndexerBase):
    def __init__(self):
        super(ChromatogramIndexer, self).__init__(
            'chromatogram', re.compile(b"<chromatogram "))


class IndexList(Sequence):

    """Wrap an arbitrary collection of :class:`TagIndexerBase`-derived
    objects, and support building XML indices.

    Attributes
    ----------
    indexers : list
        A list of :class:`TagIndexerBase`-derived objects to use when
        extracting indices from XML
    """

    def __init__(self, indexers=None):
        if indexers is None:
            indexers = []
        self.indexers = list(indexers)

    def __len__(self):
        return len([ix for ix in self.indexers if len(ix) > 0])

    def __getitem__(self, i):
        return self.indexers[i]

    def __iter__(self):
        return iter(self.indexers)

    def test(self, data, distance):
        for indexer in self:
            if indexer(data, distance):
                return True
        return False

    def __call__(self, data, distance):
        return self.test(data, distance)

    def write_index_list_xml(self, writer, distance):
        offset = distance
        n = len(self)
        with writer.element("indexList", count=str(n)):
            for index in self:
                if len(index) > 0:
                    index.write_xml(writer)
        with writer.element("indexListOffset"):
            writer.write(str(int(offset)))

    def add(self, indexer):
        self.indexers.append(indexer)


class StreamWrapperBase(object):
    '''A base class for wrapping an output stream to intercept operations.
    '''
    def __init__(self, stream):
        if isinstance(stream, basestring):
            stream = io.open(stream, 'wb')
        self.stream = stream

    if PY2:
        def write(self, b):
            return self.stream.write(b)
    else:
        def write(self, b):
            n = self.stream.write(b)
            if n is None:
                raise ValueError("The write stream %r failed to return number of bytes written" % (self.stream, ))
            return n

    def flush(self):
        return self.stream.flush()

    def close(self):
        self.stream.close()

    @property
    def closed(self):
        return self.stream.closed

    def writable(self):
        return self.stream.writable()

    @property
    def name(self):
        return self.stream.name

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if not self.closed:
            self.close()


class HashingStream(StreamWrapperBase):
    '''A write stream wrapper to compute a running SHA1 checksum as bytes are written to
    the output stream.
    '''
    def __init__(self, stream):
        super(HashingStream, self).__init__(stream)
        self._checksum = sha1()

    def write(self, b):
        n = self.stream.write(b)
        self._checksum.update(b)
        return n

    def checksum(self):
        '''Get the current checksum.

        Returns
        -------
        bytes
        '''
        return self._checksum.hexdigest()


class IndexingStream(StreamWrapperBase):
    '''An output stream wrapper that tracks running byte offsets for specific patterns.

    The output stream is also wrapped in a :class:`HashingStream`. If the innermost stream
    supports :class:`io.BufferedWriter`, this will also wrap the output stream too to reduce
    the number of calls to write through the hashing stream.
    '''

    def __init__(self, stream):
        # Make sure we keep a handle on the real hashing stream, regardless of
        # whether we wrap it in a buffering layer.
        self.hashing_stream = HashingStream(stream)
        # If the stream created supports writable and the rest of the io.IOBase
        # interface, we can combine it with io.BufferedWriter, cutting down on
        # the number of calls to write on the inner stream.
        try:
            if self.hashing_stream.writable():
                stream = io.BufferedWriter(self.hashing_stream)
            else:
                raise ValueError("Stream %s must be writable!" % self.hashing_stream.stream)
        except AttributeError:
            # No Python-level buffering
            stream = self.hashing_stream
        super(IndexingStream, self).__init__(stream)
        # A running tally of the number of bytes written
        self.accumulator = 0
        self.indices = IndexList()
        self.indices.add(SpectrumIndexer())
        self.indices.add(ChromatogramIndexer())

    # This could be optimized better, split produces a lot of copies.
    def tokenize(self, buff):
        delim = b'<'
        started_with_delim = buff.startswith(delim)
        parts = buff.split(delim)
        tail = parts[-1]
        front = parts[:-1]
        i = 0
        for part in front:
            i += 1
            if part == b"":
                continue
            if i == 1:
                if started_with_delim:
                    yield delim + part
                else:
                    yield part
            else:
                yield delim + part
        if tail.strip() and i > 0:
            yield delim + tail
        else:
            yield tail

    def write(self, data):
        n = 0
        for line in self.tokenize(data):
            self.indices.test(line, self.accumulator)
            self.accumulator += len(line)
            nl = super(IndexingStream, self).write(line)
            if nl is not None:
                n += nl
        return n

    def _raw_write(self, data):
        super(IndexingStream, self).write(data)

    def write_index(self, index, name):
        self.write("    <index name=\"{}\">\n".format(name).encode('utf-8'))
        for ref_id, index_data in index.items():
            self.write_offset(ref_id, index_data)
        self.write(b"    </index>\n")

    def checksum(self):
        # Make sure to flush the buffer into the underlying checksum stream so
        # that any final data is written.
        self.flush()
        return self.hashing_stream.checksum()

    def write_index_list(self):
        offset = self.accumulator
        self.indices.write_index_list(self, offset)

    def write_checksum(self):
        self.write(b"  <fileChecksum>")
        self.write(self.checksum().encode('utf-8'))
        self.write(b"</fileChecksum>")

    def to_xml(self, writer):
        offset = self.accumulator
        writer.flush()
        self.indices.write_index_list_xml(writer, offset)
        with writer.element("fileChecksum"):
            writer.flush()
            writer.write(self.checksum())



import warnings

from collections import defaultdict

from six import string_types as basestring

from psims.utils import ensure_iterable

from .entity import Entity
from .relationship import Relationship, Reference
from .type_definition import parse_xsdtype


synonym_scopes = {
    "EXACT",
    "BROAD",
    "NARROW",
    "RELATED"
}


def _synonym_parser(text):
    state = None
    quoted_chars = []
    scopes = []
    scope_chars = []
    references = []
    for c in text:
        if c == "\"":
            if state is None:
                state = "quote_open"
            elif state == 'quote_open':
                state = "quote_close"
            elif state == 'trailing':
                pass
            else:
                raise ValueError("Quote after quoted text!")
        else:
            if state == "quote_open":
                quoted_chars.append(c)
            elif state == "quote_close":
                if c == " ":
                    state = 'scope'
                else:
                    raise ValueError("Expected space before scope")
            elif state == 'scope':
                if c == ' ':
                    scopes.append(''.join(scope_chars))
                    scope_chars = []
                elif c == '[':
                    state = 'references'
                else:
                    scope_chars.append(c)
            elif state == 'references':
                if c != ']':
                    references.append(c)
                else:
                    state = 'trailing'
            elif state == 'trailing':
                pass
    return ''.join(quoted_chars), scopes, ''.join(references)


def synonym_parser(text):
    synonym, scopes, references = _synonym_parser(text)
    return synonym


class OBOParser(object):
    """Parser for an :title-reference:`OBO` [OBO]_ file that constructs a semantic graph.

    Attributes
    ----------
    current_term : dict
        The current term being parsed
    handle : file
        The file stream to read from
    header : defaultdict(list)
        Store the header information from the OBO file
    terms : dict
        Maps term id to :class:`~.Entity` objects

    References
    ----------
    OBO
        http://owlcollab.github.io/oboformat/doc/GO.format.obo-1_2.html
    """

    def __init__(self, handle):
        self.handle = handle
        self.terms = {}
        self.current_term = None
        self.header = defaultdict(list)
        self.parse()

    @property
    def version(self):
        try:
            return self.header['data-version']
        except KeyError:
            return None

    @property
    def name(self):
        try:
            return self.header['ontology'].upper()
        except KeyError:
            return None

    def _get_value_type(self, xref_string):
        return parse_xsdtype(xref_string)

    def pack(self):
        """Pack the currently collected OBO entry into an :class:`~.Entity`.

        Returns
        -------
        :class:`~.Entity`
        """
        if self.current_term is None:
            return
        entity = Entity(self, **{k: v[0] if len(v) == 1 else v for k, v in self.current_term.items()})
        self._expand_is_a(entity)
        self._expand_relationship(entity)
        self._expand_synonym(entity)
        self._expand_xref(entity)
        self._expand_property_value(entity)
        self.terms[entity['id']] = entity
        self.current_term = None

    def _expand_is_a(self, entity):
        if "is_a" in entity.data:
            is_as = entity['is_a']
            if isinstance(is_as, basestring):
                is_as = Reference.fromstring(is_as)
            else:
                is_as = list(map(Reference.fromstring, is_as))
            entity['is_a'] = is_as

    def _expand_relationship(self, entity):
        if 'relationship' in entity.data:
            relationships = entity['relationship']
            if not isinstance(relationships, list):
                relationships = [relationships]
            relationships = [Relationship.fromstring(r) for r in relationships]
            for rel in relationships:
                entity.setdefault(rel.predicate, [])
                entity[rel.predicate].append(rel)

    def _expand_synonym(self, entity):
        if 'synonym' in entity.data:
            synonyms = entity['synonym']
            if not isinstance(synonyms, list):
                synonyms = [synonyms]
            synonyms = list(map(synonym_parser, synonyms))
            entity['synonym'] = synonyms

    def _expand_xref(self, entity):
        entity.value_type = None
        if 'xref' in entity.data:
            xref = entity['xref']
            if isinstance(xref, basestring):
                xref = [xref]
            for x in xref:
                key, value = x.split(":", 1)
                if key == 'value-type':
                    entity.value_type = self._get_value_type(x)
                else:
                    if value.startswith("\""):
                        value, dtype = value.rsplit(" ", 1)
                        dtype = parse_xsdtype(dtype)
                        if dtype is not None:
                            value = dtype(value[1:-1])
                    entity[key] = value

    def _expand_property_value(self, entity):
        if "property_value" in entity.data:
            for prop_val in ensure_iterable(entity.data['property_value']):
                prop, val = prop_val.split(" ", 1)
                prop = prop.strip(": ")
                val = val.strip()
                if val.startswith("\""):
                    val, dtype = val.rsplit(" ", 1)
                    dtype = parse_xsdtype(dtype)
                    try:
                        val = dtype(val[1:-1])
                    except (ValueError, TypeError):
                        pass
                entity[prop] = val

    def _connect_parents(self):
        """Walk the semantic graph up the parent hierarchy, binding child
        to parent through ``is_a`` :class:`~.Reference` connections.
        """
        for term in self.terms.values():
            try:
                if isinstance(term.is_a, Reference):
                    self.terms[term.is_a].children.append(term)
                else:
                    for is_a in term.is_a:
                        self.terms[is_a].children.append(term)
            except KeyError:
                continue

    def _pack_if_occupied(self):
        if self.current_term is not None:
            self.pack()

    def _simplify_header_information(self):
        self.header = {
            k: v if len(v) > 1 else v[0] for k, v in self.header.items()
        }

    def parse(self):
        """Iteratively parse a binary file stream for an OBO file into a
        semantic graph.
        """
        in_header = True
        for line in self.handle.readlines():
            line = line.decode('utf-8')
            line = line.strip()
            if not line:
                in_header = False
                continue
            elif in_header:
                key, val = line.split(":", 1)
                self.header[key].append(val.strip())
            elif line == "[Typedef]":
                self._pack_if_occupied()
                self.current_term = None
            elif line == "[Term]":
                self._pack_if_occupied()
                self.current_term = defaultdict(list)
            else:
                if self.current_term is None:
                    continue
                key, sep, val = line.partition(":")
                self.current_term[key].append(val.strip())
        self.pack()
        self._connect_parents()
        self._simplify_header_information()
        self.handle.close()

    def __getitem__(self, key):
        return self.terms[key]

    def __iter__(self):
        return iter(self.terms.items())


class OBOWriter(object):
    def __init__(self, stream):
        self.stream = stream

    def write_header(self, header):
        for key, value in header:
            if isinstance(value, (list, tuple)):
                for v in value:
                    self.stream.write("%s: %s\n" % (key, v))
            else:
                self.stream.write("%s: %s\n" % (key, value))
        self.stream.write("\n")
        self.stream.write("\n")

    def write_term(self, term):
        self.stream.write("[Term]\nid: %s\nname: %s\ndef: \"%s\"\n" %
                    (term.id, term.name, term.definition))
        for xref in term.get('xref', []):
            self.stream.write("xref: ")
        for is_a in ensure_iterable(term.get("is_a", [])):
            self.stream.write("is_a: %s" % str(is_a))
        seen = set()
        for syn in term.get('synonyms', []):
            if syn in seen:
                continue
            seen.add(syn)
            self.stream.write("synonym: \"%s\" EXACT\n" % str(syn).replace("\n", "\\n"))
        for prop in term.get('property_value', []):
            self.stream.write("property_value: %s\n" % prop)
        self.stream.write("\n")

    def write_vocabulary(self, vocabulary):
        self.write_header(vocabular.metadata)
        for term in vocabulary.terms.values():
            self.write_term(term)



import re
import warnings

from contextlib import contextmanager
from collections import deque, OrderedDict

from lxml import etree

from six import string_types as basestring, add_metaclass, text_type, PY3

from . import controlled_vocabulary
from .controlled_vocabulary import obj_to_xsdtype
from .validation import validate


def make_counter(start=1):
    '''
    Create a functor whose only internal piece of data is a mutable container
    with a reference to an integer, `start`. When the functor is called, it returns
    current `int` value of `start` and increments the mutable value by one.

    Parameters
    ----------
    start: int, optional
        The number to start counting from. Defaults to `1`.

    Returns
    -------
    int:
        The next number in the count progression.
    '''
    start = [start]

    def count_up():
        '''A closure returning incrementally larger integers'''
        ret_val = start[0]
        start[0] += 1
        return ret_val
    return count_up


def camelize(name):
    """Adapts an attribute name from "snake_case" to "camelCase"
    to make lookups on Element.attrib easier.

    Parameters
    ----------
    name : str
        Attribute name

    Returns
    -------
    str
        transformed name
    """
    parts = name.split("_")
    if len(parts) > 1:
        return ''.join([parts[0]] + [part.title() if part != "ref" else "_ref" for part in parts[1:]])
    else:
        return name


def id_maker(type_name, id_number):
    '''Generate a consistent ID that is unique within a document,
    assuming `id_number` is unique within the tag type.
    '''
    return "%s_%s" % (type_name.upper(), str(id_number))


def sanitize_id(string):
    """Remove characters from a string which would be invalid
    in XML identifiers

    Parameters
    ----------
    string : str

    Returns
    -------
    str
    """
    string = re.sub(r"\s", '_', string)
    string = re.sub(r"\\|/", '', string)
    return string


NO_TRACK = object()


class ElementType(type):
    """A metaclass to keep a count of the number of times
    an instance of each derived class is created.
    """
    _cache = {}

    def __new__(mcs, name, parents, attrs):
        new_type = type.__new__(mcs, name, parents, attrs)
        tag_name = attrs.get("tag_name")
        if attrs.get("_track") is NO_TRACK:
            return new_type
        if not hasattr(mcs, "_cache"):
            mcs._cache = dict()
        mcs._cache[name] = new_type
        if tag_name is not None:
            mcs._cache[tag_name] = new_type
        return new_type


def attrencode(o):
    """A simple function to convert most
    basic python types to a string form
    which is safe to serialize in XML
    attributes

    Parameters
    ----------
    o : object
        Attribute value to encode

    Returns
    -------
    str:
        The encoded value
    """
    if isinstance(o, bool):
        return text_type(o).lower()
    else:
        return text_type(o)


@add_metaclass(ElementType)
class TagBase(object):
    """Represent a single XML element with arbitrary attributes.

    Mocks the :class:`Mapping` interface

    Attributes
    ----------
    attrs : dict
        The attributes of the element. :meth:`__getattr__` falls back
        to querying :attr:`attrs`, trying both requested name and ``camelCase``
        versions of the name
    is_open : bool
        Whether the element has been opened as a context manager
    tag_name : str
        The name of the element
    text : str
        The body text of the element
    type_attrs : dict
        Attributes common to all elements of this type
    id : str
        The @id attribute of the element.
    """

    type_attrs = {}

    def __init__(self, tag_name=None, text="", **attrs):
        self.tag_name = tag_name or self.tag_name
        _id = attrs.pop('id', None)
        self.attrs = {}
        self.attrs.update(self.type_attrs)
        self.text = text
        self.attrs.update(attrs)
        # When passing through a XMLWriterMixin.element() call, tags may be reconstructed
        # and any set ids will be passed through the attrs dictionary, but the `with_id`
        # flag won't be propagated. `_force_id` preserves this.
        self._force_id = True
        if _id is None:
            self._id_number = None
            self._id_string = None
            self._force_id = False
        elif isinstance(_id, int):
            self._id_number = _id
            self._id_string = None
        else:
            self._id_number = None
            self._id_string = _id
        self.is_open = False

    def __getattr__(self, key):
        try:
            return self.attrs[key]
        except KeyError:
            try:
                return self.attrs[camelize(key)]
            except KeyError:
                raise AttributeError("%s has no attribute %s" % (self.__class__.__name__, key))

    # Support Mapping Interface
    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def keys(self):
        if self.id is not None:
            yield "id"
        for key in self.attrs:
            yield key

    def get(self, name, default):
        return getattr(self, name, default)

    @property
    def id(self):
        if self._id_string is None and self._id_number is not None:
            self._id_string = id_maker(self.tag_name, self._id_number)
        return self._id_string

    def element(self, xml_file=None, with_id=False):
        """Create an XML element using either a materialized :class:`lxml.etree.Element`
        if ``xml_file`` is :const:`None` or the ephemeral context manager element
        :class:`lxml.etree._FileWriterElement`.

        The element will be constructed with all attributes in :attr:`attrs` where the
        value is not :const:`None`.

        Parameters
        ----------
        xml_file : :class:`XMLWriterMixin`, optional
            The XML writer to build the element for
        with_id : bool, optional
            Whether to require the ID attribute be present and rendered

        Returns
        -------
        :class:`lxml.etree.Element` or :class:`lxml.etree._FileWriterElement`
            Description

        Raises
        ------
        ValueError
            Description
        """
        with_id = with_id or self._force_id
        attrs = {k: attrencode(v) for k, v in self.attrs.items() if v is not None}
        if with_id:
            if self.id is None:
                raise ValueError("Required id for %r but id was None" % (self,))
            attrs['id'] = self.id
        if xml_file is None:
            elt = etree.Element(self.tag_name, attrs)
            if self.text:
                elt.text = self.text
            return elt
        else:
            return xml_file.element(self.tag_name, **attrs)

    def write(self, xml_file, with_id=False):
        """Write this element to file

        Parameters
        ----------
        xml_file : :class:`XMLWriterMixin`
            The XML writer to build the element for
        with_id : bool, optional
            Whether to require the ID attribute be present and rendered

        See Also
        --------
        :meth:`element`
        """
        el = self.element(with_id=with_id)
        xml_file.write(el)

    def bind(self, xml_file):
        self._xml_file = xml_file

    @contextmanager
    def begin(self, xml_file, with_id=False):
        with self.element(xml_file, with_id):
            self.is_open = True
            yield
        self.is_open = False

    def __call__(self, xml_file=None, with_id=False):
        """Alias of :meth:`element`
        """
        return self.element(xml_file, with_id)

    def __repr__(self):
        return "<%s id=\"%s\" %s>" % (self.tag_name, self.id, " ".join("%s=\"%s\"" % (
            k, attrencode(v)) for k, v in self.attrs.items()))

    def __eq__(self, other):
        try:
            return self.attrs == other.attrs
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self.attrs != other.attrs
        except AttributeError:
            return True

    def __hash__(self):
        return hash((self.tag_name, frozenset(self.attrs.items())))


def identity(x):
    return x


def _make_tag_type(name, **attrs):
    """Creates a new TagBase-derived class dynamically at runtime.
    The new type will be cached.

    Parameters
    ----------
    name : str
        The tag name
    **attrs : dict
        Any class-wide attributes to include

    Returns
    -------
    type
        A TagBase subclass
    """
    return type(name, (TagBase,), {"tag_name": name, "type_attrs": attrs})


def _element(_tag_name, *args, **kwargs):
    """Construct a subclass instance of :class:`TagBase` with the given
    tag name. All other arguments are forwarded to the :class:`TagBase`
    constructor

    Parameters
    ----------
    _tag_name : str
        The name of the tag type to create
    *args
        Arbitrary arguments for the tag
    **kwargs
        Key word arguments for the tag

    Returns
    -------
    :class:`TagBase`
    """

    try:
        eltype = ElementType._cache[_tag_name]
    except KeyError:
        eltype = _make_tag_type(_tag_name)
    return eltype(*args, **kwargs)


def element(xml_file, _tag_name, *args, **kwargs):
    """Construct and immediately write a subclass instance of
    :class:`TagBase` with the given tag name. All other arguments
    are forwarded to the :class:`TagBase` constructor

    Parameters
    ----------
    xml_file : :class:`XMLWriterMixin`
        The XML writer to write to
    _tag_name : str
        The name of the tag type to create
    *args
        Arbitrary arguments for the tag
    **kwargs
        Key word arguments for the tag
    """
    with_id = kwargs.pop("with_id", False)
    if isinstance(_tag_name, basestring):
        el = _element(_tag_name, *args, **kwargs)
    else:
        el = _tag_name
    return el.element(xml_file=xml_file, with_id=with_id)


class AttrProperty(object):
    def __init__(self, name, transform=None):
        if transform is None:
            transform = self._default_transform
        self.name = name
        self.transform = transform

    def _default_transform(self, value):
        return value

    def __get__(self, inst, cls=None):
        return inst.attrs.get(self.name)

    def __set__(self, inst, value):
        inst.attrs[self.name] = self.transform(value)

    def __delete__(self, inst):
        try:
            inst.attrs.pop(self.name)
        except KeyError:
            pass


class XSDTypingProperty(AttrProperty):

    def __get__(self, inst, cls=None):
        value = inst.attrs.get(self.name)
        xsdtype = obj_to_xsdtype(value)
        if xsdtype:
            inst.attrs['type'] = xsdtype
        return value

    def __set__(self, inst, value):
        inst.attrs[self.name] = self.transform(value)
        value = inst.attrs[self.name]
        xsdtype = obj_to_xsdtype(value)
        if xsdtype:
            inst.attrs['type'] = xsdtype



class CVParam(TagBase):
    """Represents a ``<cvParam />``

    .. note::
        This element holds additional data or annotation. Only controlled values
        are allowed here

    Attributes
    ----------
    name: str
        The human-readable name of the parameter
    accession: str
        The within-vocabulary unique identifier of the parameter
    value: object
        The value of the parameter, to be converted to text

    """

    tag_name = "cvParam"
    _track = NO_TRACK

    @classmethod
    def param(cls, name, value=None, **attrs):
        if isinstance(name, cls):
            return name
        elif isinstance(name, (tuple, list)):
            name, value = name
        else:
            if value is None:
                return cls(name=name, **attrs)
            else:
                return cls(name=name, value=value, **attrs)

    @staticmethod
    def _normalize_units(attrs):
        if 'unit_cv_ref' in attrs:
            attrs["unitCvRef"] = attrs.pop("unit_cv_ref")
        if 'unit_accession' in attrs:
            attrs['unitAccession'] = attrs.pop("unit_accession")
        if 'unit_name' in attrs:
            attrs['unitName'] = attrs.pop("unit_name")

        return attrs

    def __init__(self, accession=None, name=None, ref=None, value=None, **attrs):
        if ref is not None:
            attrs["cvRef"] = ref
        if accession is not None:
            attrs["accession"] = accession
        if name is not None:
            attrs["name"] = name
        if value is not None:
            attrs['value'] = value
        else:
            attrs['value'] = ''

        attrs = self._normalize_units(attrs)

        super(CVParam, self).__init__(self.tag_name, **attrs)
        self.patch_accession(accession, ref)

    value = AttrProperty("value")
    ref = AttrProperty("cvRef")
    name = AttrProperty("name")
    accession = AttrProperty("accession")

    unit_name = AttrProperty("unitName")
    unit_accession = AttrProperty("unitAccession")
    unit_cv_ref = AttrProperty("unitCvRef")


    def __call__(self, *args, **kwargs):
        self.write(*args, **kwargs)

    def __repr__(self):
        return "<%s %s>" % (self.tag_name, " ".join("%s=\"%s\"" % (
            k, str(v)) for k, v in self.attrs.items()))

    def patch_accession(self, accession, ref):
        if accession is not None:
            if isinstance(accession, int):
                accession = "%s:%d" % (ref, accession)
                self.attrs['accession'] = accession
            else:
                self.attrs['accession'] = accession


class UserParam(CVParam):
    """Represents a ``<userParam />`` element

    .. note::
        Uncontrolled user parameters (essentially allowing free text). Before
        using these, one should verify whether there is an appropriate CV term
        available, and if so, use the CV term instead
    """

    tag_name = "userParam"
    accession = None

    value = XSDTypingProperty('value')
    type = AttrProperty('type')

    def __init__(self, accession=None, name=None, ref=None, value=None, **attrs):
        super(UserParam, self).__init__(
            accession=accession, name=name, ref=ref, value=value, **attrs)
        # get XSDTypingProperty to force property __get__
        self.value


class ParamGroupReference(TagBase):
    tag_name = "referenceableParamGroupRef"

    def __init__(self, ref):
        self.ref = ref
        super(ParamGroupReference, self).__init__(self.tag_name, ref=ref)

    def __call__(self, *args, **kwargs):
        self.write(*args, **kwargs)


class CVCollection(object):
    """A partially unique collection of :class:`CV` objects.
    """
    def __init__(self, cvs=None):
        self.storage = OrderedDict()
        if cvs:
            self.update(cvs)

    def add(self, cv):
        """Add `cv` to the collection.

        If the :attr:`CV.id` is aleady present in the collection, a warning
        will be issued.

        Parameters
        ----------
        cv : :class:`CV`
            The controlled vocabulary reference
        """
        if cv.id in self.storage:
            old = self.storage[cv.id]
            warnings.warn(
                "Replacing {cv.id} {old.version}@{old.uri} with {cv.version}@{cv.uri}".format(
                    cv=cv, old=old))
        self.storage[cv.id] = cv
        return self

    def update(self, collection):
        """Add each element of `collection` to `self`, calling :meth:`add` on each
        element.

        Parameters
        ----------
        collection : :class:`~.Iterable`
            Any iterable collection of :class:`CV`

        See Also
        --------
        :meth:`add`
        """
        for cv in collection:
            self.add(cv)

    append = add
    extend = update

    def copy(self):
        """Create a copy of `self`

        Returns
        -------
        :class:`CVCollection`
        """
        return self.__class__(self)

    def __add__(self, other):
        copy = self.copy()
        copy.update(other)
        return copy

    def __iadd__(self, other):
        self.update(other)
        return self

    def __getitem__(self, key):
        return self.storage[key]

    def __iter__(self):
        return iter(self.storage.values())

    def __len__(self):
        return len(self.storage)

    def __repr__(self):
        template = "{self.__class__.__name__}({members})"
        members = list(self)
        return template.format(self=self, members=members)


class CV(object):
    """Represent a Controlled Vocabulary associated with the current document.

    The controlled vocabulary referenced must specify a URI that will be used
    to either download the definitions from, to be matched to a cache of pre-built
    vocabularies, or to be matched with a set of special resolution rules for
    handled by the :attr:`resolver`.

    This object acts as a lazy-loading proxy for
    :class:`~.controlled_vocabulary.ControlledVocabulary`.

    Attributes
    ----------
    full_name : str
        The full name of the controlled vocabulary, which may or may not
        be identical to the :attr:`id`
    id : str
        A short unique identifier for the controlled vocabulary
    options : :class:`dict`
        Additional information that may be used during resolution
    resolver : :class:`~.VocabularyResolver`
        The resolver which will handle all requests for this controlled vocabulary
    uri : str
        The location where the definition can be found
    version : str
        The version of the vocabulary resolved
    vocabulary : :class:`~.ControlledVocabulary`
        The parsed term graph defining this vocabulary
    """

    def __init__(self, full_name, id, uri, version=None, resolver=None, **kwargs):
        self.full_name = full_name
        self.id = id
        self.uri = uri
        self._version = version
        self.options = kwargs
        self._vocabulary = None
        self.resolver = None

    def __hash__(self):
        return hash(self.uri)

    def match(self, other):
        return self.uri == other.uri and self.id == other.id and self.full_name == other.full_name

    def __eq__(self, other):
        return self.match(other) and self.version == other.version

    def __ne__(self, other):
        return not self == other

    @property
    def version(self):
        if self._version is None:
            try:
                self._version = self.vocabulary.version
            except AttributeError:
                pass
        return self._version

    @property
    def vocabulary(self):
        if self._vocabulary is None:
            self._vocabulary = self.load()
        return self._vocabulary

    def load(self, handle=None):
        """Load the vocabulary definition from source

        Assumes that the definition is in OBO format

        Parameters
        ----------
        handle : file-like, optional
            An optional file handle which can be read from directly.

        Returns
        -------
        :class:`~.ControlledVocabulary`
            The parsed vocabulary

        Raises
        ------
        KeyError
            When all fallback mechanisms fail, a KeyError is raised
        """
        resolver = self.resolver or controlled_vocabulary.obo_cache
        if handle is None:
            try:
                fp = resolver.resolve(self.uri)
                cv = controlled_vocabulary.ControlledVocabulary.from_obo(fp)
            except ValueError:
                fp = resolver.fallback(self.uri)
                if fp is not None:
                    cv = controlled_vocabulary.ControlledVocabulary.from_obo(fp)
                else:
                    raise KeyError(self.uri)
        else:
            cv = controlled_vocabulary.ControlledVocabulary.from_obo(handle)
        try:
            cv.id = self.id
        except Exception:
            import traceback
            traceback.print_exc()
        return cv

    def __getitem__(self, key):
        return self.vocabulary[key]

    def query(self, *args, **kwargs):
        return self.vocabulary.query(*args, **kwargs)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.full_name!r}, {self.id!r}, {self.uri!r}, {version!r})"
        version = self._version
        if version is None:
            version = "?"
        return template.format(self=self, version=version)


class ProvidedCV(CV):
    """A wrapper around another object that provides the same basic interface
    as :class:`CV` from that object, provided through the :attr:`converter`
    function

    Attributes
    ----------
    converter : Callable
        A function that converts elements of the provided vocabulary into
        something matching the :class:`~.controlled_vocabulary.Entity` interface.
    """

    def __init__(self, id, uri, converter=identity, **kwargs):
        self.converter = converter
        super(ProvidedCV, self).__init__(id=id, uri=uri, **kwargs)

    def load(self, handle=None):
        resolver = self.resolver or controlled_vocabulary.obo_cache
        try:
            cv = resolver.resolve(self.uri)
        except ValueError:
            cv = resolver.fallback(self.uri)
            if cv is None:
                raise LookupError(self.uri)
        try:
            cv.id = self.id
        except Exception:
            pass
        return cv

    def __getitem__(self, key):
        return self.converter(super(ProvidedCV, self).__getitem__(key), self)


class XMLWriterMixin(object):
    """A mixin class to provide methods for writing
    XML elements and aggregate Components.

    Attributes
    ----------
    verbose : bool
        Controls debug printing
    writer: lxml.etree._IncrementalFileWriter
        The low-level XML writer used by :attr:`xmlfile`
    """
    verbose = False

    @contextmanager
    def element(self, element_name, **kwargs):
        """Construct and immediately open a subclass instance of
        :class:`TagBase` with the given tag name. All other arguments
        are forwarded to the :class:`TagBase` constructor.

        Parameters
        ----------
        element_name: str
            The name of the tag type to create
        *args
            Arbitrary arguments for the tag
        **kwargs
            Key word arguments for the tag

        See Also
        --------
        :func:`element`
        """
        if self.verbose:
            print("In XMLWriterMixin.element", element_name, kwargs)
        try:
            if isinstance(element_name, basestring):
                with element(self.writer, element_name, **kwargs):
                    yield
            else:
                with element_name.element(self.writer, **kwargs):
                    yield
        except AttributeError:
            if self.writer is None:
                raise ValueError(
                    "This writer has not yet been created."
                    " Make sure to use this object as a context manager using the "
                    "`with` notation or by explicitly calling its __enter__ and "
                    "__exit__ methods.")
            else:
                raise

    def write(self, *args, **kwargs):
        """Either write a complete XML sub-tree or add free text to the file stream

        Parameters
        ----------
        arg: str or :class:`lxml.etree.Element`
            The entity to be written out.
        """
        if self.verbose:
            print(args[0])
        try:
            self.writer.write(*args, **kwargs)
        except AttributeError:
            if self.writer is None:
                raise ValueError(
                    "This writer has not yet been created."
                    " Make sure to use this object as a context manager using the "
                    "`with` notation or by explicitly calling its __enter__ and "
                    "__exit__ methods.")
            else:
                raise

    def flush(self):
        self.writer.flush()


class XMLFormattingStreamWriter(object):
    """Wraps a writable stream with a layer emulating the
     class:`lxml.etree.xmlfile` interface, save that it automatically
     formats and indents the XML generated without requiring a separate
     pass through the XML pretty printing processing

    Attributes
    ----------
    indent_chars : str
        The characters to indent with
    indent_level : int
        The current indentation level
    stream : :class:`io.IOBase`
        The stream to wrap
    writer : :class:`lxml.etree._IncrementalFileWriter`
        The actual XML writer
    wrote_text_stack : :class:`collections.deque`
        A stack to track if a layer wrote text in one of its children and should not
        have its end tag written on a new line
    xmlfile : :class:`lxml.etree.xmlfile`
        The actual XML writer's controller
    """

    def __init__(self, stream, encoding=None, indent='  ', **kwargs):
        self.stream = stream
        self.xmlfile = etree.xmlfile(self.stream, encoding=encoding, buffered=False, **kwargs)
        self.writer = None
        self.indent_level = 0
        self.indent_chars = indent
        self.wrote_text_stack = deque()

    def __enter__(self):
        self.writer = self.xmlfile.__enter__()
        return self

    def __exit__(self, *args):
        self.xmlfile.__exit__(*args)

    def _indent_tag(self):
        self.writer.write('\n' + (self.indent_chars * self.indent_level))

    @contextmanager
    def element(self, *args, **kwargs):
        if self.indent_level > 0:
            self._indent_tag()
        elt = self.writer.element(*args, **kwargs)
        elt.__enter__()
        self.wrote_text_stack.append(False)
        self.indent_level += 1
        yield
        self.indent_level -= 1
        if not self.wrote_text_stack[-1]:
            self._indent_tag()
        elt.__exit__(None, None, None)
        self.wrote_text_stack.pop()

    def write(self, *args, **kwargs):
        if isinstance(args[0], basestring if not PY3 else (str, bytes)):
            self.wrote_text_stack[-1] = True
        if not self.wrote_text_stack[-1]:
            if self.indent_level > 0:
                self._indent_tag()
        self.writer.write(*args, **kwargs)

    def flush(self):
        self.writer.flush()

    def write_doctype(self):
        self.writer.write_doctype()

    def write_declaration(self):
        self.writer.write_declaration()


class XMLDocumentWriter(XMLWriterMixin):
    """A base class for types which are used to
    write complete XML documents.

    Attributes
    ----------
    outfile : file
        A writable file object
    toplevel : TagBase
        The top-level XML tag
    xmlfile : lxml.etree.xmlfile
        The XML formatter
    writer : lxml.etree._IncrementalFileWriter
        The low-level XML writer used by :attr:`xmlfile`
    """
    @staticmethod
    def toplevel_tag():
        """Overridable method to construct the appropriate
        tag for :attr:`toplevel`

        Returns
        -------
        TagBase
        """
        return _element("xmldocument")

    def __init__(self, outfile, close=False, encoding=None, **kwargs):
        if encoding is None:
            encoding = 'utf-8'
        self.outfile = outfile
        self.encoding = encoding
        self.xmlfile = XMLFormattingStreamWriter(outfile, encoding=encoding, **kwargs)
        self._writer = None
        self.toplevel = None
        self._close = close

    def _should_close(self):
        if self._close is None:
            return hasattr(self.outfile, 'close')
        return bool(self._close)

    @property
    def writer(self):
        if self._writer is None:
            raise ValueError(
                "The writer hasn't been started yet. Either use the object as a context manager"
                " or call the begin() method")
        return self._writer

    @writer.setter
    def writer(self, value):
        self._writer = value

    def begin(self):
        """Writes the doctype and starts the low-level writing machinery
        """
        if self._has_begun():
            return
        self.writer = self.xmlfile.__enter__()
        self.writer.write_declaration()
        toplevel = self.toplevel_tag()
        if isinstance(toplevel, TagBase):
            toplevel_element = element(self.writer, toplevel)
        else:
            toplevel_element = toplevel
        self.toplevel = toplevel_element
        self.toplevel.__enter__()

    def _has_begun(self):
        return self.toplevel is not None

    def __enter__(self):
        """Begins writing, opening the top-level tag
        """
        self.begin()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Closes the top-level tag, the XML formatter,
        and the file itself.
        """
        self.end(exc_type, exc_value, traceback)

    def end(self, exc_type=None, exc_value=None, traceback=None):
        """Ends the XML document, and flushes and closes the file
        """
        self.toplevel.__exit__(exc_type, exc_value, traceback)
        self.writer.flush()
        try:
            self.xmlfile.__exit__(exc_type, exc_value, traceback)
        except etree.SerialisationError as err:
            # There seems to be a bug when `__exit__`ing xmlfile objects, usually
            # when they get large see here: https://bugs.launchpad.net/lxml/+bug/1570388
            # https://github.com/mobiusklein/psims/issues/7
            if str(err).startswith('unknown error'):
                warnings.warn("Error closing file: {}".format(err))
            else:
                raise
        try:
            self.flush()
        except Exception:
            pass
        if self._should_close():
            self.close()

    def controlled_vocabularies(self):
        """Write out the `<cvList>` element and all its children,
        including both this format's default controlled vocabularies
        and those passed as arguments to this method.this

        This method requires writing to have begun.
        """
        cvlist = self.CVList(self.vocabularies)
        cvlist.write(self.writer)

    def close(self):
        try:
            self.outfile.close()
        except AttributeError:
            pass

    def flush(self):
        try:
            self.outfile.flush()
        except AttributeError:
            self.writer.flush()

    def format(self, outfile=None):
        """This method is deprecated. Previously, the serialization
        process did not indent the XML in-place and the lxml pretty printer
        had to be invoked separately. With the addition of :class:`XMLFormattingStreamWriter`,
        the XML stream is formatted in-place as it is being streamed to file.
        """
        warnings.warn(
            "This method is no longer necessary, XML is formatted automatically",
            DeprecationWarning, 2)

    def validate(self):
        """Attempt to perform XSD validation on the XML document
        this writer wrote

        Returns
        -------
        bool:
            Whether or not the document was valid

        lxml.etree.XMLSchema:
            The schema object where errors are logged

        Raises
        ------
        TypeError
            When the file cannot be recovered from the writer object,
            a :class:`TypeError` is thrown
        """
        prev = None
        try:
            if isinstance(self.outfile, basestring):
                fname = self.outfile
            else:
                fname = self.outfile.name
        except AttributeError:
            if hasattr(self.outfile, 'seek'):
                prev = self.outfile.tell()
                self.outfile.seek(0)
                fname = self.outfile
            else:
                raise TypeError("Can't get file from %r" % (self.outfile,))
        result, schema = validate(fname)
        if prev is not None:
            self.outfile.seek(prev)
        return result, schema



import pytest

from io import BytesIO

from psims import document
from psims.mzml import writer, components
from .utils import output_path


def test_repr_borrow():
    buffer = BytesIO()
    f = writer.MzMLWriter(buffer)

    with f:
        f.controlled_vocabularies()
        assert repr(f.SourceFile) == repr(components.SourceFile)
    f.close()


def test_referential_integrity():
    ctx = document.DocumentContext()
    ctx["Spam"][1] = document.id_maker("Spam", 1)
    assert ctx['Spam'][1] == "SPAM_1"
    with pytest.warns(document.ReferentialIntegrityWarning):
        assert ctx['Spam'][2] == "SPAM_2"
    with pytest.warns(document.ReferentialIntegrityWarning):
        assert ctx['Spam']['With Purple Eggs'] == 'With Purple Eggs'


def test_xmlwriter(output_path):
    f = writer.MzMLWriter(output_path)
    with f:
        pass
    f.close()
    try:
        f.format()
        with open(output_path, 'rb') as fh:
            print(fh.readline())
    except OSError:
        pass

    buffer = BytesIO()
    f = writer.MzMLWriter(buffer)
    with f:
        f.write("Spam")
    f.format()
    f.close()
    with open(output_path, 'rb') as fh:
        print(fh.readline())



import re

from pyteomics import mzid
from pyteomics.auxiliary import cvquery

from psims.utils import ensure_iterable
from psims.mzid import MzIdentMLWriter

from .utils import TransformerBase


N = float('inf')
K = 1000


def identity(x):
    return x


class MzIdentMLParser(mzid.MzIdentML):
    def _handle_param(self, element, **kwargs):
        try:
            element.attrib["value"]
        except KeyError:
            element.attrib["value"] = ""
        return super(MzIdentMLParser, self)._handle_param(element, **kwargs)

    def reset(self):
        super(MzIdentMLParser, self).reset()
        self.seek(0)


class MzIdentMLTranslater(TransformerBase):
    def _uncamel(self, name):
        temp = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', temp).lower()

    def _translate_keys(self, d, keys):
        for key in keys:
            try:
                d[self._uncamel(key)] = d.pop(key)
            except KeyError:
                continue

    def _extract_params(self, d):
        params = []
        if isinstance(d, str):
            params.append({
                "name": d, "accession": getattr(d, 'accession')
            })
            return params
        for key, value in list(d.items()):
            if hasattr(key, 'accession'):
                accession = key.accession
                value_type = identity
                if accession:
                    term = self.writer.term(accession)
                    try:
                        if term.value_type is not None:
                            value_type = term.value_type
                    except KeyError:
                        pass
                if isinstance(value, list):
                    value_list = value
                    for value in value_list:
                        unit = None
                        if hasattr(value, 'unit_info'):
                            unit = value.unit_info
                        try:
                            cast_value = value_type(value)
                        except ValueError:
                            cast_value = value
                        param = {
                            "name": key, "value": cast_value
                        }
                        if accession:
                            param['accession'] = accession
                        if unit:
                            param['unit_name'] = unit
                        params.append(param)
                else:
                    unit = None
                    if hasattr(value, 'unit_info'):
                        unit = value.unit_info
                    if value_type is not None:
                        try:
                            cast_value = value_type(value)
                        except ValueError:
                            cast_value = value
                    else:
                        cast_value = value
                    param = {
                        "name": key, "value": cast_value
                    }
                    if accession:
                        param['accession'] = accession
                    if unit:
                        param['unit_name'] = unit
                    params.append(param)

                d.pop(key)
        return params

    def __init__(self, input_stream, output_stream):
        self.input_stream = input_stream
        self.output_stream = output_stream
        self.reader = MzIdentMLParser(input_stream, retrieve_refs=False, iterative=True)
        self.writer = MzIdentMLWriter(output_stream, version=self.reader.version_info[0])

    def insert_software_record(self):
        # self.writer.AnalysisSoftware(
        #     name='psims-example-MzIdentMLTransformer',
        #     id='psims-example-MzIdentMLTransformer').write(self.writer.writer)
        pass

    def copy_provenance(self):
        self.log("Copying Provenance Information")
        self.reader.reset()
        try:
            analysis_software_list = next(self.reader.iterfind("AnalysisSoftwareList"))
        except StopIteration:
            analysis_software_list = {}

        analysis_software_list = map(
            self.format_analysis_software, analysis_software_list.get("AnalysisSoftware", []))

        with self.writer.AnalysisSoftwareList(analysis_software_list):
            self.insert_software_record()

        try:
            self.reader.reset()
            provider = self.format_provider(next(self.reader.iterfind("Provider")))
            provider.write(self.writer)
        except StopIteration:
            pass

        try:
            self.reader.reset()
            audit_collection = self.format_autit_collection(
                next(self.reader.iterfind("AuditCollection")))
            audit_collection.write(self.writer.writer)
        except StopIteration:
            pass

    def format_analysis_software(self, software):
        d = {}
        d.update(software)
        d['name'] = d.pop('SoftwareName')
        contact_role = ensure_iterable(d.pop("ContactRole", {}))[0]
        d['role'] = contact_role.get("Role")
        d['contact'] = contact_role.get('contact_ref')
        return self.writer.AnalysisSoftware.ensure(d)

    def format_provider(self, provider):
        d = {}
        d.update(provider)
        contact_role = ensure_iterable(d.pop("ContactRole", {}))[0]
        d['role'] = contact_role.get("Role")
        d['contact'] = contact_role.get('contact_ref')
        return self.writer.Provider.ensure(d)

    def format_person(self, person):
        d = {}
        d.update(person)
        self._translate_keys(d, ['firstName', 'lastName', 'midInitials'])
        d['affiliations'] = [a.get('organization_ref') for a in ensure_iterable(d.pop("Affiliation", {}))
                             if a.get('organization_ref') is not None]
        return self.writer.Person.ensure(d)

    def format_organization(self, organization):
        d = {}
        d.update(organization)
        d['parent'] = d.pop("Parent", {}).get("organization_ref")
        return self.writer.Organization.ensure(d)

    def format_autit_collection(self, audit_collection):
        organization = map(self.format_organization, ensure_iterable(audit_collection.pop("Organization", None)))
        person = map(self.format_person, ensure_iterable(audit_collection.pop("Person", None)))
        return self.writer.AuditCollection(person, organization)

    def copy_sequence_collection(self):
        self.log("Copying Sequence Collection")
        writer = self.writer
        reader = self.reader
        with writer.sequence_collection():
            reader.reset()
            i = 0
            for db_seq in map(self.format_db_sequence, reader.iterfind("DBSequence")):
                i += 1
                db_seq.write(self.writer)
                if i % K == 0:
                    self.log("Copied %d DBSequences" % i)
                if i > N:
                    break

            reader.reset()
            i = 0
            for peptide in map(self.format_peptide, reader.iterfind("Peptide")):
                i += 1
                peptide.write(self.writer)
                if i % K == 0:
                    self.log("Copied %d Peptides" % i)
                if i > N:
                    break

            reader.reset()
            i = 0
            for peptide_ev in map(self.format_peptide_evidence, reader.iterfind("PeptideEvidence")):
                i += 1
                peptide_ev.write(self.writer)
                if i % K == 0:
                    self.log("Copied %d PeptideEvidence" % i)
                if i > N:
                    break

    def format_db_sequence(self, db_sequence):
        d = dict(db_sequence)
        d['search_database_id'] = d.pop("searchDatabase_ref", None)
        d['sequence'] = d.pop("Seq", None)
        return self.writer.DBSequence.ensure(d)

    def format_peptide(self, peptide):
        d = dict(peptide)
        d['peptide_sequence'] = d.pop("PeptideSequence")
        d['modifications'] = map(self.format_modification, d.pop("Modification", []))
        d['substitutions'] = map(self.format_substitution, d.pop("SubstitutionModification", []))
        return self.writer.Peptide.ensure(d)

    def format_modification(self, mod):
        temp = dict(mod)
        d = dict()
        d['location'] = temp.pop("location", None)
        d['monoisotopic_mass_delta'] = temp.pop("monoisotopicMassDelta", None)
        d['avg_mass_delta'] = temp.pop("avgMassDelta", None)
        d['residues'] = temp.pop("residues", None)
        term_dict = cvquery(temp)
        crosslinking_donor_or_receiver = None
        has_identity = False
        params = []
        d['params'] = params

        for key, value in list(term_dict.items()):
            term = self.writer.term(key)
            if term.is_of_type('UNIMOD:0'):
                d['name'] = term.name
                has_identity = True
                term_dict.pop(key)
            elif term.is_of_type("MS:1001460"):
                d['name'] = value
                d['accession'] = "MS:1001460"
                term_dict.pop(key)
                has_identity = True
            # crosslinking donor
            elif term.is_of_type("MS:1002508"):
                crosslinking_donor_or_receiver = {"name": term.name, 'accession': term.id, 'value': int(value)}
                term_dict.pop(key)
            elif term.is_of_type("XLMOD:00001") or term.is_of_type("XLMOD:00002"):
                d['name'] = term.name
                has_identity = True
                term_dict.pop(key)

        if not has_identity and crosslinking_donor_or_receiver:
            d.update(crosslinking_donor_or_receiver)
        elif crosslinking_donor_or_receiver:
            params.append(crosslinking_donor_or_receiver)
        return self.writer.Modification.ensure(d)

    def format_substitution(self, sub):
        d = dict(sub)
        self._translate_keys(d, ["originalResidue", "replacementResidue", "monoisotopicMassDelta"])
        return self.writer.SubstitutionModification.ensure(d)

    def format_peptide_evidence(self, evidence):
        d = dict(evidence)
        d['peptide_id'] = d.pop("peptide_ref")
        d['db_sequence_id'] = d.pop("dBSequence_ref")
        d['start_position'] = d.pop('start', None)
        d['end_position'] = d.pop('end', None)
        d['translation_table_id'] = d.pop('translationTable_ref', None)
        d['is_decoy'] = d.pop("isDecoy", False)
        return self.writer.PeptideEvidence.ensure(d)

    def copy_analysis_collection(self):
        reader = self.reader
        writer = self.writer
        reader.reset()
        analysis_collection = next(reader.iterfind("AnalysisCollection"))
        with self.writer.analysis_collection():
            for si in analysis_collection.get("SpectrumIdentification", []):
                self.format_spectrum_identification(si).write(writer)
            for pi in ensure_iterable(analysis_collection.get("ProteinDetection", [])):
                self.format_protein_detection(pi).write(writer)

    def format_enzyme(self, enz):
        d = dict(enz)
        d['name'] = d.pop("EnzymeName", None)
        d['site_regexp'] = d.pop('SiteRegexp', None)
        d['missed_cleavages'] = d.pop("missedCleavages", 0)
        self._translate_keys(d, ["minDistance", "semiSpecific", "nTermGain", "cTermGain"])
        return self.writer.Enzyme.ensure(d)

    def format_search_modification(self, mod):
        temp = dict(mod)
        d = dict()
        d['fixed'] = temp.pop("fixedMod", False)
        d['mass_delta'] = temp.pop("massDelta")
        d['specificity'] = temp.pop('SpecificityRules', [])
        d['residues'] = temp.pop("residues")
        term_dict = cvquery(temp)
        crosslinking_donor_or_receiver = None
        has_identity = False
        params = []
        d['params'] = params

        for key, value in list(term_dict.items()):
            term = self.writer.term(key)
            if term.is_of_type('UNIMOD:0'):
                d['name'] = term.name
                has_identity = True
                term_dict.pop(key)
            elif term.is_of_type("MS:1001460"):
                d['name'] = value
                d['accession'] = "MS:1001460"
                term_dict.pop(key)
                has_identity = True
            # crosslinking donor
            elif term.is_of_type("MS:1002508"):
                crosslinking_donor_or_receiver = {"name": term.name, 'accession': term.id, 'value': int(value)}
                term_dict.pop(key)
            elif term.is_of_type("XLMOD:00001") or term.is_of_type("XLMOD:00002"):
                d['name'] = term.name
                has_identity = True
                term_dict.pop(key)

        if not has_identity and crosslinking_donor_or_receiver:
            d.update(crosslinking_donor_or_receiver)
        elif crosslinking_donor_or_receiver:
            params.append(crosslinking_donor_or_receiver)
        # return d
        return self.writer.SearchModification.ensure(d)

    def format_tolerance(self, tol, tp):
        if not tol:
            return None
        term_dict = cvquery(tol)
        value = term_dict.get('MS:1001413')
        low = {"accession": 'MS:1001413', "value": value, 'unit_name': getattr(value, 'unit_info', None)}
        value = term_dict.get('MS:1001412')
        high = {"accession": 'MS:1001412', "value": value, 'unit_name': getattr(value, 'unit_info', None)}
        return tp(low, high)

    def format_spectrum_identification_protocol(self, sip):
        d = dict(sip)
        d['additional_search_params'] = d.pop("AdditionalSearchParams", [])
        enzymes = d.pop("Enzymes", {})
        d['enzymes'] = self.writer.Enzymes(
            map(self.format_enzyme, ensure_iterable(enzymes.get("Enzyme", []))),
            independent=enzymes.get("independent"))
        d['modification_params'] = map(
            self.format_search_modification, ensure_iterable(
                d.pop('ModificationParams', {}).get("SearchModification", [])))
        d['search_type'] = d.pop("SearchType", None)
        d['parent_tolerance'] = self.format_tolerance(d.pop("ParentTolerance", {}), self.writer.ParentTolerance)
        d['fragment_tolerance'] = self.format_tolerance(d.pop("FragmentTolerance", {}), self.writer.FragmentTolerance)
        d['threshold'] = self._extract_params(d.pop("Threshold"))
        d['analysis_software_id'] = d.pop("analysisSoftware_ref", None)
        return self.writer.SpectrumIdentificationProtocol.ensure(d)

    def format_protein_detection_protocol(self, pdp):
        d = dict(pdp)
        d['threshold'] = self._extract_params(d.pop("Threshold"))
        d['analysis_software_id'] = d.pop("analysisSoftware_ref", None)
        d['params'] = d.pop("AnalysisParams", [])
        return self.writer.ProteinDetectionProtocol.ensure(d)

    def copy_analysis_protocol_collection(self):
        self.log("Copying Protocols")
        self.reader.reset()
        apc = next(self.reader.iterfind("AnalysisProtocolCollection", retrieve_refs=False))
        protocols = []
        for sip in ensure_iterable(apc.get("SpectrumIdentificationProtocol", [])):
            protocols.append(self.format_spectrum_identification_protocol(sip))
        for pdp in ensure_iterable(apc.get("ProteinDetectionProtocol", [])):
            protocols.append(self.format_protein_detection_protocol(pdp))
        with self.writer.analysis_protocol_collection():
            for protocol in protocols:
                protocol.write(self.writer)

    def format_spectrum_identification(self, si):
        d = dict(si)
        self._translate_keys(d, ['activityDate'])
        d['spectrum_identification_list_id'] = d.pop("spectrumIdentificationList_ref")
        d['spectrum_identification_protocol_id'] = d.pop("spectrumIdentificationProtocol_ref")
        d['spectra_data_ids_used'] = [s['spectraData_ref'] for s in d.pop('InputSpectra', [])]
        d['search_database_ids_used'] = [s['searchDatabase_ref'] for s in d.pop('SearchDatabaseRef', [])]
        return self.writer.SpectrumIdentification.ensure(d)

    def format_protein_detection(self, pi):
        d = dict(pi)
        self._translate_keys(d, ['activityDate'])
        d['protein_detection_list_id'] = d.pop("proteinDetectionList_ref")
        d['protein_detection_protocol_id'] = d.pop("proteinDetectionProtocol_ref")
        d['spectrum_identification_ids_used'] = [
            si['spectrumIdentificationList_ref'] for si in d.pop("InputSpectrumIdentifications", [])
        ]
        return self.writer.ProteinDetection.ensure(d)

    def copy_inputs(self):
        self.log("Copying Inputs")
        self.reader.reset()
        inputs = next(self.reader.iterfind('Inputs'))
        source_files = map(self.format_source_file, ensure_iterable(inputs.get("SourceFile")))
        search_databases = map(self.format_search_database, ensure_iterable(inputs.get("SearchDatabase")))
        spectra_data = map(self.format_spectra_data, ensure_iterable(inputs.get("SpectraData")))
        self.writer.inputs(source_files, search_databases, spectra_data)

    def format_file_record(self, fr):
        d = dict(fr)
        fileformat = d.pop("FileFormat", {})
        format_type = self._extract_params(fileformat)
        if format_type:
            d['fileformat'] = format_type[0]
        d['externalformat'] = d.pop("ExternalFormatDocumentation", None)
        return d

    def format_source_file(self, sf):
        d = self.format_file_record(sf)
        return self.writer.SourceFile.ensure(d)

    def format_spectra_data(self, sd):
        d = self.format_file_record(sd)
        self._translate_keys(d, ['SpectrumIDFormat'])
        return self.writer.SpectraData.ensure(d)

    def format_search_database(self, sd):
        d = self.format_file_record(sd)
        d['name'] = d.pop("DatabaseName", None)
        if isinstance(d['name'], dict):
            try:
                d['name'] = list(d['name'])[0]
            except KeyError:
                pass
        self._translate_keys(d, ["numDatabaseSequences", "numResidues", "releaseDate", "version"])
        return self.writer.SearchDatabase.ensure(d)

    def copy_analysis_data(self):
        self.log("Copying Analysis Data")
        reader = self.reader
        writer = self.writer
        with writer.analysis_data():
            reader.reset()
            for spectrum_identification_list in reader.iterfind("SpectrumIdentificationList"):
                list_id = spectrum_identification_list.pop('id')
                with writer.spectrum_identification_list(list_id):
                    i = 0
                    for spectrum_id_result in spectrum_identification_list.pop("SpectrumIdentificationResult", []):
                        i += 1
                        result = self.format_spectrum_identification_result(spectrum_id_result)
                        if result is None:
                            continue
                        with result:
                            for item in map(
                                    self.format_spectrum_identification_item,
                                    spectrum_id_result['SpectrumIdentificationItem']):
                                if item is None:
                                    continue
                                item.write(self.writer)
                        if i % K == 0:
                            self.log("Copied %d SpectrumIdentificationResults" % i)
                        if i > N:
                            break

            reader.reset()
            for protein_detection_list in reader.iterfind("ProteinDetectionList"):
                list_id = protein_detection_list.pop('id')
                size = len(protein_detection_list['ProteinAmbiguityGroup'])
                with writer.protein_detection_list(list_id, count=size):
                    i = 0
                    for pag in protein_detection_list.pop('ProteinAmbiguityGroup'):
                        i += 1
                        result = self.format_protein_ambiguity_group(pag)
                        if result is None:
                            continue
                        with result:
                            j = 0
                            for prot in pag['ProteinDetectionHypothesis']:
                                result = self.format_protein_detection_hypothesis(prot)
                                if result is None:
                                    continue
                                with result:
                                    k = 0
                                    for pept in prot['PeptideHypothesis']:
                                        result = self.format_peptide_detection_hypothesis(pept)
                                        if result is None:
                                            continue
                                        result.write(self.writer)
                                    k += 1
                                    if k > N:
                                        break
                            j += 1
                            if j > N:
                                break
                        if i % K == 0:
                            self.log("Copied %d ProteinAmbiguityGroups" % i)
                        if i > N:
                            break

    def format_spectrum_identification_result(self, sir):
        d = dict(sir)
        self._translate_keys(d, ['spectrumID'])
        d['spectra_data_id'] = d.pop('spectraData_ref', None)
        d.pop('SpectrumIdentificationItem', [])
        d['params'] = self._extract_params(d)
        return self.writer.SpectrumIdentificationResult.ensure(d)

    def format_spectrum_identification_item(self, sii):
        d = dict(sii)
        self._translate_keys(
            d, ('chargeState', 'passThreshold', 'experimentalMassToCharge', 'calculatedMassToCharge'))
        d['peptide_id'] = d.pop("peptide_ref")
        d['peptide_evidence_ids'] = [p['peptideEvidence_ref'] for p in d.pop("PeptideEvidenceRef", [])]
        d['params'] = self._extract_params(d)
        return self.writer.SpectrumIdentificationItem.ensure(d)

    def format_protein_ambiguity_group(self, pag):
        d = dict(pag)
        d.pop("ProteinDetectionHypothesis", [])
        d['protein_detection_hypotheses'] = []
        d['params'] = self._extract_params(d)
        return self.writer.ProteinAmbiguityGroup.ensure(d)

    def format_protein_detection_hypothesis(self, prot):
        d = dict(prot)
        d['db_sequence_id'] = d.pop("dBSequence_ref")
        d.pop("PeptideHypothesis", [])
        d['peptide_hypotheses'] = []
        self._translate_keys(d, ['passThreshold', ])
        d['params'] = self._extract_params(d)
        return self.writer.ProteinDetectionHypothesis.ensure(d)

    def format_peptide_detection_hypothesis(self, pept):
        d = dict(pept)
        d['peptide_evidence_id'] = d.pop("peptideEvidence_ref")
        d['spectrum_identification_ids'] = [
            siir['spectrumIdentificationItem_ref'] for siir in d.pop('SpectrumIdentificationItemRef')]
        d['params'] = self._extract_params(d)
        return self.writer.PeptideHypothesis.ensure(d)

    def write(self):
        writer = self.writer
        with writer:
            writer.controlled_vocabularies()
            self.copy_provenance()
            self.copy_sequence_collection()
            self.copy_analysis_collection()
            self.copy_analysis_protocol_collection()
            with writer.data_collection():
                self.copy_inputs()
                self.copy_analysis_data()



import base64
import zlib

import numpy as np

import six

if six.PY2:
    decode_base64 = base64.decodestring
else:
    decode_base64 = base64.decodebytes


COMPRESSION_NONE = 'none'
COMPRESSION_ZLIB = 'zlib'
COMPRESSION_LINEAR = 'linear'
COMPRESSION_DELTA = 'delta'


encoding_map = {
    32: np.float32,
    64: np.float64,
    '64-bit integer': np.int64,
    'MS:1000522': np.int64,
    'MS:1000519': np.int32,
    '32-bit integer': np.int32,
    'MS:1000520': np.float16,
    '16-bit float': np.float16,
    'MS:1000521': np.float32,
    '32-bit float': np.float32,
    'MS:1000523': np.float64,
    '64-bit float': np.float64,
    'MS:1001479': np.bytes_,
    'null-terminated ASCII string': np.bytes_,
    float: np.float64,
    int: np.int32,
    None: np.float32
}


dtype_to_encoding = {
    np.int64: '64-bit integer',
    np.int32: '32-bit integer',
    np.float16: '16-bit float',
    np.float32: '32-bit float',
    np.float64: '64-bit float',
    np.bytes_: 'null-terminated ASCII string',
}


compression_map = {
    COMPRESSION_ZLIB: "zlib compression",
    COMPRESSION_NONE: 'no compression',
    None: 'no compression',
    False: 'no compression',
    True: "zlib compression",
    COMPRESSION_LINEAR: "linear prediction",
    COMPRESSION_DELTA: "delta prediction"
}


for dtype in list(encoding_map.values()):
    encoding_map[dtype] = dtype


def encode_array(array, compression=COMPRESSION_NONE, dtype=np.float32):
    if compression == COMPRESSION_LINEAR or compression == COMPRESSION_DELTA:
        array = np.asanyarray(array).astype(dtype, copy=False)
        if compression == COMPRESSION_LINEAR:
            array = linear_encode(array, copy=True)
        else:
            array = delta_encode(array, copy=True)
        bytestring = array.tobytes()
        # Now don't apply any additional encoding to the generated bytestring
        compression = COMPRESSION_NONE
    else:
        bytestring = np.asanyarray(array).astype(dtype, copy=False).tobytes()
    if compression == COMPRESSION_NONE:
        bytestring = bytestring
    elif compression == COMPRESSION_ZLIB:
        bytestring = zlib.compress(bytestring)
    else:
        raise ValueError("Unknown compression: %s" % compression)
    encoded_string = base64.standard_b64encode(bytestring)
    return encoded_string


def encode_array_direct(array, compression=COMPRESSION_NONE, dtype=np.float32):
    array = np.asanyarray(array).astype(dtype, copy=False)
    if compression == COMPRESSION_LINEAR:
        array = linear_encode(array, copy=True)
    elif compression == COMPRESSION_DELTA:
        array = delta_encode(array, copy=True)
    return array


def decode_array(bytestring, compression=COMPRESSION_NONE, dtype=np.float32):
    try:
        decoded_string = bytestring.encode("ascii")
    except AttributeError:
        decoded_string = bytestring
    decoded_string = decode_base64(decoded_string)
    if compression == COMPRESSION_NONE:
        decoded_string = decoded_string
    elif compression == COMPRESSION_ZLIB:
        decoded_string = zlib.decompress(decoded_string)
    else:
        raise ValueError("Unknown compression: %s" % compression)
    array = np.frombuffer(decoded_string, dtype=dtype)
    return array


def delta_predict(data, copy=True):
    '''Reverse the lossy transformation of the delta compression
    helper.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        The data to transform
    copy : bool
        Whether to make a copy of the data array or transform it in-place.

    Returns
    -------
    :class:`numpy.ndarray`
        The transformed data array
    '''
    if copy:
        out = data.copy()
    else:
        out = data
    for i in range(2, len(data)):
        out[i] = out[i] + out[i - 1] - out[0]
    return out


def linear_predict(data, copy=True):
    '''Reverse the lossy transformation of the linear interpolation compression
    helper.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        The data to transform
    copy : bool
        Whether to make a copy of the data array or transform it in-place.

    Returns
    -------
    :class:`numpy.ndarray`
        The transformed data array
    '''
    if copy:
        out = data.copy()
    else:
        out = data
    for i in range(2, len(data)):
        out[i] = out[i] + 2 * out[i - 1] - out[i - 2] - out[1]
    return out


def linear_encode(data, copy=True):
    if copy:
        out = data.copy()
    else:
        out = data
    n = len(data)
    if n < 3:
        return data
    prev2 = out[0]
    prev1 = out[1]
    offset = out[1]
    for i in range(2, n):
        out[i] = offset + out[i] - 2 * prev1 + prev2
        tmp = prev1
        prev1 = out[i] + 2 * prev1 - prev2 - offset
        prev2 = tmp
    return out


def delta_encode(data, copy=True):
    if copy:
        out = data.copy()
    else:
        out = data
    n = len(data)
    if n < 2:
        return data
    prev = out[0]
    offset = out[0]
    for i in range(1, n):
        tmp = out[i]
        out[i] = offset + out[i] - prev
        prev = tmp
    return out



from collections import deque
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from psims.utils import ensure_iterable

from .type_definition import parse_xsdtype


class Entity(Mapping):
    def __init__(self, vocabulary=None, **attributes):
        self.data = dict(attributes)
        self.children = []
        self.vocabulary = vocabulary

    def get(self, key, default=None):
        return self.data.get(key, default)

    def __contains__(self, key):
        return key in self.data

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key in ("vocabulary", "children", "data"):
            object.__setattr__(self, key, value)
        else:
            self[key] = value

    def __dir__(self):
        keys = set(object.__dir__(self)) | set(self.keys())
        return sorted(keys)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.keys())

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    def setdefault(self, key, value):
        self.data.setdefault(key, value)

    @property
    def definition(self):
        return self.data.get("def", '')

    @definition.setter
    def definition(self, value):
        self.data['def'] = value

    def parent(self):
        try:
            reference = self.is_a
        except KeyError:
            return None
        try:
            return self.vocabulary[reference]
        except TypeError:
            return [self.vocabulary[r] for r in reference]

    def __repr__(self):
        template = 'Entity({self.id!r}, {self.name!r}, {self.definition!r})'
        return template.format(self=self)

    def is_of_type(self, tp):
        try:
            tp = self.vocabulary[tp]
        except KeyError:
            return False
        stack = deque([self])
        while stack:
            ref = stack.pop()
            if ref == tp:
                return True
            stack.extend(ensure_iterable(ref.parent()))
        return False



from six import add_metaclass


class ParamManagingProperty(object):
    '''A descriptor that will delegate value storage to
    the bound instance's :attr:`binding`, and can be
    enumerated by :class:`ParamManagingMeta` to set up
    defaults.

    Attributes
    ----------
    name: str
        The key to delegate to.
    default_value: object
        The value to initialize the bound key to, or a callable
        to provide that value.
    aliases: dict
        Alternative names to map the name
    '''

    def __init__(self, name, default_value=None, aliases=None):
        self.name = name
        self.default_value = default_value
        self.aliases = {a: name for a in aliases or []}

    def __get__(self, obj, cls=None):
        try:
            value = obj.binding[self.name]
        except KeyError:
            value = obj.binding[self.name] = self.default_value
        return value

    def __set__(self, obj, value):
        obj.binding[self.name] = value

    def __delete__(self, obj):
        del obj.binding[self.name]

    def _init_default(self, obj):
        if callable(self.default_value):
            obj.binding.setdefault(self.name, self.default_value())
        else:
            obj.binding.setdefault(self.name, self.default_value)

    def __repr__(self):
        return "{self.__class__.__name__}({self.name!r}, {self.default_value!r})".format(self=self)


class ParamManagingMeta(type):
    '''Register :class:`ParamManagingProperty` bound to
    the initializing class so they can be enumerated
    at a later point.

    Attributes
    ----------
    _binding_props: list
        A list of :class:`ParamManagingProperty` attached to this type
    '''

    def __new__(cls, name, parents, attrs):
        new_type = type.__new__(cls, name, parents, attrs)
        new_type._binding_props = []
        for key, value in attrs.items():
            if isinstance(value, ParamManagingProperty):
                new_type._binding_props.append(value)
        for parent in parents:
            new_type._binding_props.extend(
                getattr(parent, '_binding_props', []))
        return new_type


@add_metaclass(ParamManagingMeta)
class ElementBuilder(object):
    '''A type to accumulate information in to allow
    '''
    binding_name = None

    def __init__(self, source, binding=None, params=None, **kwargs):
        if binding is None:
            binding = dict()
        else:
            binding = dict(binding)

        self.source = source
        self.binding = dict()
        self._init_defaults()
        self.update(binding)
        self.update(kwargs)

    def add_param(self, value):
        self.params.append(value)
        return self

    def set(self, key=None, value=None, **kwargs):
        if key is not None:
            self[key] = value
        if kwargs:
            for key, value in kwargs.items():
                self[key] = value
        return self

    params = ParamManagingProperty('params', list)

    def _init_defaults(self):
        aliases = dict()
        for prop in self._binding_props:
            prop._init_default(self)
            aliases.update(prop.aliases)
        self._aliases = aliases

    def __repr__(self):
        fmt = "{self.__class__.__name__}({self.binding})"
        return fmt.format(self=self)

    def keys(self):
        return self.binding.keys()

    def values(self):
        return self.binding.values()

    def items(self):
        return self.binding.items()

    def get(self, key, value=None):
        if key in self._aliases:
            key = self._aliases[key]
        return self.binding.get(key, value)

    def __getitem__(self, key):
        if key in self._aliases:
            key = self._aliases[key]
        return self.binding[key]

    def __setitem__(self, key, item):
        if key in self._aliases:
            key = self._aliases[key]
        self.binding[key] = item
        return self

    def __iter__(self):
        return iter(self.binding)

    def update(self, _value):
        for key, value in _value.items():
            self[key] = value
        return self

    def pack(self):
        result = {}
        for key, value in self.binding.items():
            if isinstance(value, ElementBuilder):
                if value.binding_name is not None:
                    key = value.binding_name
                value = value.pack()
            if isinstance(value, list):
                value = [v.pack() if isinstance(v, ElementBuilder)
                         else v for v in value]
            if key == 'params' and not value:
                continue
            result[key] = value
        return result
